import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

import msgtools.parser.cpp.language as cpplanguage
import msgtools.parser.simulink.language as simlanguage

cpplanguage.namespace = "<MSGFULLNAME>_"
cpplanguage.firstParam = "m_data"
cpplanguage.firstParamDecl = "uint8_t* m_data"
cpplanguage.const = ""
cpplanguage.enumNamespace = 1
enumNamespace = 1
cpplanguage.functionPrefix = "INLINE "
cpplanguage.enumClass = ""

#enums = cpplanguage.enums
accessors = cpplanguage.accessors
#declarations = cpplanguage.declarations
#structPacking = cpplanguage.structPacking
#structUnpacking = cpplanguage.structUnpacking
initCode = cpplanguage.initCode
#get_fields = simlanguage.get_fields
#set_fields = simlanguage.set_fields

def enums(e):
    ret = ""
    for enum in e:
        ret +=  "typedef enum {\n"
        for option in enum["Options"]:
            optionName = OptionName(option)
            if enumNamespace != 0:
                optionName = "<MSGFULLNAME>"+"_"+enum["Name"] + "_" + optionName
            ret += "    " + optionName + " = "+str(option["Value"]) + ",\n"
        ret = ret[:-2] + "\n"
        ret += "} <MSGFULLNAME>" + "_" + enum["Name"] + ";\n"
    return ret

def declarations(msg, msg_enums):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    retType = cpplanguage.fieldType(field)
                    if fieldHasConversion(bits):
                        retType = typeForScaledInt(bits)
                    elif "Enum" in bits:
                        retType = "<MSGFULLNAME>" + "_" + bits["Enum"]
                    if MsgParser.fieldCount(field) == 1:
                        ret.append(retType + " " + bits["Name"] + ";")
                    else:
                        ret.append(retType + " " + bits["Name"] + "["+str(MsgParser.fieldCount(field))+"];")
            else:
                retType = cpplanguage.fieldType(field)
                if fieldHasConversion(field):
                    retType = typeForScaledInt(field)
                elif "Enum" in field:
                    retType = "<MSGFULLNAME>" + "_" + field["Enum"]
                if MsgParser.fieldCount(field) == 1:
                    ret.append(retType + " " + field["Name"] + ";")
                else:
                    ret.append(retType + " " + field["Name"] + "["+str(MsgParser.fieldCount(field))+"];")
    return ret


def set_field(msg, field, count):
    if "Enum" in field:
        enumCast = "(<MSGFULLNAME>Message" + "::" + field["Enum"] + ")"
    else:
        enumCast = ""

    if count == 1:
        ret = "msg->Set%s(%sunpackedMsg->%s);" % (field["Name"], enumCast, field["Name"])
    else:
        ret = ""
        for idx in range(count):
            ret = ret + "msg->Set%s(%sunpackedMsg->%s[%d], %d);\n" % (field["Name"], enumCast, field["Name"], idx, idx)
        ret = ret[:-1]
    return ret


def set_fields(msg):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    ret.append(set_field(msg, bitfield, fieldCount(field)))
            else:
                ret.append(set_field(msg, field, fieldCount(field)))
    return ret

def get_field(msg, field, count):
    if "Enum" in field:
        enumCast = "(<MSGFULLNAME>" + "_" + field["Enum"] + ")"
    else:
        enumCast = ""

    if count == 1:
        ret = "unpackedMsg->%s = %smsg->Get%s();" % (field["Name"], enumCast, field["Name"])
    else:
        ret = ""
        for idx in range(count):
            ret = ret + "unpackedMsg->%s[%d] = %smsg->Get%s(%d);\n" % (field["Name"], idx, enumCast, field["Name"], idx)
        ret = ret[:-1]
    return ret


def get_fields(msg):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    ret.append(get_field(msg, bitfield, fieldCount(field)))
            else:
                ret.append(get_field(msg, field, fieldCount(field)))

    return ret



from PyQt5 import QtCore, QtWidgets

from msgtools.lib.messaging import Messaging

class DialogFieldInfo:
    def __init__(self, fieldInfo):
        self.fieldInfo = fieldInfo
        if fieldInfo.type == "enumeration":
            self.widget = QtWidgets.QComboBox()
            self.widget.addItems(fieldInfo.enum[0].keys())
            # there's some odd behavior in the UI when the box is editable :(
            # if you want it editable, uncomment this line, and play around and see if you like it
            #self.widget.setEditable(1)
            # store a hash table of value->ComboBoxIndex
            # this is NOT the same as value->enumIndex!
            self.comboBoxIndexOfEnum = {}
            for i in range(0, self.widget.count()):
                self.comboBoxIndexOfEnum[self.widget.itemText(i)] = i
        else:
            self.widget = QtWidgets.QLineEdit()
    def setValueInMsg(self, msg):
        #TODO ought to handle arrays, bitfields, enums
        if self.fieldInfo.type == "enumeration":
            value = self.widget.currentText()
        else:
            value = self.widget.text()
        Messaging.set(msg, self.fieldInfo, value)

# dialog box to let user enter fields of a message and send it
class CreateMsgDialog(QtWidgets.QDialog):
    def __init__(self, parent, msgClass):
        super(CreateMsgDialog, self).__init__(parent)
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.setWindowTitle(msgClass.MsgName())
        
        # construct a message to store data into
        self.msg = msgClass()
        
        layout = QtWidgets.QFormLayout()
        self.setLayout(layout)

        self.headerFields = []
        for fieldInfo in Messaging.hdr.fields:
            if fieldInfo.bitfieldInfo:
                for bitfieldInfo in fieldInfo.bitfieldInfo:
                    if bitfieldInfo.idbits == 0 and bitfieldInfo.name != "DataLength" and bitfieldInfo.name != "Time":
                        field = DialogFieldInfo(bitfieldInfo)
                        self.headerFields.append(field)
                        layout.addRow(QtWidgets.QLabel(bitfieldInfo.name), field.widget)
            else:
                if fieldInfo.idbits == 0 and fieldInfo.name != "DataLength" and fieldInfo.name != "Time":
                    field = DialogFieldInfo(fieldInfo)
                    self.headerFields.append(field)
                    layout.addRow(QtWidgets.QLabel(fieldInfo.name), field.widget)
        
        self.msgFields = []
        for fieldInfo in msgClass.fields:
            field = DialogFieldInfo(fieldInfo)
            self.msgFields.append(field)
            layout.addRow(QtWidgets.QLabel(fieldInfo.name), field.widget)

        buttonBox = QtWidgets.QDialogButtonBox(self)
        #buttonBox.setOrientation(QtCore.Qt.Horizontal)
        buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        buttonBox.accepted.connect(self.constructMessage)
        buttonBox.rejected.connect(self.reject)
        layout.addRow(buttonBox)

    def constructMessage(self):
        for field in self.headerFields:
            field.setValueInMsg(self.msg.hdr)
        for field in self.msgFields:
            field.setValueInMsg(self.msg)
        self.accept()



import copy
import datetime
import fileinput
import gevent
import importlib
import json
import logging
import os
import signal
import sys

from msgtools.console.client import Client
from msgtools.lib.messaging import Messaging
from msgtools.sim.sim_exec import SimExec, StyleAdapter
from msgtools.lib.message import Message

# Script files can have blank lines and comments (started with #)
# Commands within them are of the form:
# [dt] cmd [option]
# where dt is optional, and "option" is required for some but not all commands.
# If dt is present, the script line execution will not occur until dt seconds have
# elapsed since the previous script line execution.
class ScriptReader:
    # This is the count of how many options each command has.
    # Any commands in the script file that are not in the dict below will cause
    # the sim to exit with an error message.
    OPTION_COUNT_OF_CMD = {"exit": 0, "send": 1, "delay": 1, "sleep": 1, "set": 1, "load": 1}
    class Line:
        def __init__(self, next_line, file, logging):
            next_line = next_line.strip()
            self.lineinfo = "%s:%d [%s]" % (file.filename(), file.lineno(), next_line)

            # try splitting into three: DT CMD OPTION
            # if that fails, split into two: CMD OPTION
            next_index = 0
            try:
                parts = next_line.split(' ', maxsplit = 2)
                might_be_dt = parts[0]
                self.exec_time = float(might_be_dt)
                next_index += 1
            except ValueError:
                parts = next_line.split(' ', maxsplit = 1)
                self.exec_time = 0.0

            # try getting the command as lower case text
            try:
                self.command = parts[next_index].lower()
            except ValueError:
                logging.error("Error parsing %s, invalid command" % (self.lineinfo))
                exit(1)

            if not self.command in ScriptReader.OPTION_COUNT_OF_CMD.keys():
                logging.error("Invalid command [%s] in script line %s" % (self.command, self.lineinfo))
                exit(1)

            option_count = ScriptReader.OPTION_COUNT_OF_CMD[self.command]
            if option_count > 0:
                next_index += 1
                try:
                    self.option = parts[next_index]
                    if self.command in ["sleep", "delay"]:
                        if self.exec_time == 0.0:
                            self.exec_time = float(self.option)
                except IndexError:
                    logging.error("Error parsing %s, invalid options" % (self.lineinfo))
                    exit(1)

    def __init__(self, filename, start_time, logging):
        self.logging = logging
        # _script_files is a list, so we can have scripts load other scripts,
        # and when the child script is done, the parent script resumes.
        self._script_files = []
        self._script_files.append(fileinput.FileInput(filename))
        self._last_time = start_time
        self._read_line()

    def get_next_command(self, t):
        # return the next script input unless it's not time to send the next one
        while len(self._script_files)>0 and t >= self._line.exec_time:
            if self._line.command == "load":
                self.logging.warning("Loading script %s" % (self._line.option))
                self._script_files.append(fileinput.FileInput(self._line.option))
                self._read_line()
            else:
                ret = copy.deepcopy(self._line)
                self._read_line()
                return ret
        return None

    def _read_line(self):
        while 1:
            # when there's no files left, return
            if len(self._script_files) == 0:
                self._line = None
                return
            next_line = self._script_files[-1].readline()
            # When we get to the end of the file, pop it off the list
            if not next_line:
                self._script_files.pop()
                continue

            # if the line we read is blank or a comment, continue to the next line.
            if next_line.isspace():
                continue
            if next_line.startswith("#"):
                continue
            break
        # remove leading and trailing spaces, and then split on first space
        self._line = ScriptReader.Line(next_line, self._script_files[-1], self.logging)
        self._line.exec_time += self._last_time
        self._last_time = self._line.exec_time

# The sim base class executes a FDM (flight dynamics model), and optionally
# FSW (flight software), along with an optional test script (text based, or
# python).  The various entities are all run according to simulation time,
# and can pass messages amongst themselves and the network.
class SimBaseClass:
    def __init__(self, dt, fsw_dir, fsw_module, args):
        SimExec.sim_exists = True
        self._args = args
        logging.basicConfig(level=10*self._args.loglevel, format='')
        self.logging = StyleAdapter(logging.getLogger(__name__), self)
        self.time_stats = SimExec(dt, self._args, self.logging)
        self.cxn = Client('Sim', timeout=0.0)

        self.log_file_type = None
        self.log_file = None
        if self._args.log != None:
            if self._args.log == "":
                self._args.log = "json"
            self.start_log(self._args.log)

        # Open the initial script file
        if self._args.script != None:
            script_filename = self._args.script[0]
            if script_filename.endswith('.py'):
                script_args = self._args.script[1:]
                #print("script_args {%s}" % (script_args))
                # spawn the script using gevent
                spec = importlib.util.spec_from_file_location("sim.script", script_filename)
                sim_script = importlib.util.module_from_spec(spec)
                sys.modules["sim.script"] = sim_script
                spec.loader.exec_module(sim_script)

                #TODO does something need to join with this?!
                gevent.spawn(sim_script.main, *script_args)
                self.script_reader = None
            else:
                self.script_reader = ScriptReader(script_filename, self.get_time(), self.logging)
        else:
            self.script_reader = None

        if self._args.fsw:
            # go up until we find the FSW dir, then add it to the path, and
            # import the FSW module.
            search_dir = os.getcwd()
            while not os.path.isdir(os.path.join(search_dir, fsw_dir)):
                last_search_dir = search_dir
                search_dir = os.path.abspath(os.path.join(search_dir, os.pardir))
                if last_search_dir == search_dir:
                    exit("ERROR!  Cannot find FSW path %s upstream of %s" % (fsw_dir, os.getcwd()))
            fsw_dir = os.path.join(search_dir, fsw_dir)
            self.logging.info("Adding %s for path, to import FSW python module" % (fsw_dir))
            sys.path.append(fsw_dir)
            self.fsw = importlib.import_module(fsw_module)
        else:
            self.fsw = None

        gevent.signal_handler(signal.SIGTERM, SimBaseClass.close, signal.SIGTERM)
        gevent.signal_handler(signal.SIGINT, SimBaseClass.close, signal.SIGINT)

        if self._args.asap:
            if self._args.hitl:
                print("ERROR!")
                print("The simulation was run with --asap and --hitl, and these are incompatible.")
                print("You must choose only one of --asap or --hitl.")
                exit(1)
            if not self._args.fsw:
                print("ERROR!")
                print("The simulation was run with --asap but without --fsw.")
                print("If you want to run --asap, you also need --fsw to run the flight software internal to the sim")
                exit(1)

        if not Client._socket_connected:
            if not self._args.fsw and not self._args.hitl:
                print("ERROR!")
                print("The simulation was run without --fsw and without --hitl, but has been unable to make a socket connection to msgserver!")
                print("This means the sim has nothing to talk to!")
                print("You probably should run with one of:")
                print("1) --fsw to run the flight software internal to the sim")
                print("2) --hitl to use hardware links or virtual hardware links to send simulation data to the flight software")
                print("3) or you should run msgserver and run")
                print("  a. the flight software as a Linux software application on this PC")
                print("  b. the flight software on an embedded processor with a link to msgserver (USB, CAN, serial, etc.)")
                if self._args.asap:
                    print("Note that the sim will not make a socket connection to msgserver when run with --asap,")
                    print("so if you want to use msgserver, you should NOT specify --asap")
                exit(1)
            if not self._args.log:
                print("ERROR!")
                print("The simulation was run without --log, but has been unable to make a socket connection to msgserver!")
                print("You ought to either specify --log so data is recorded for post-processing, or else have msgserver")
                print("running so you can use other applications to see what's happening or log it.")
                print("If you don't include --log or run msgserver, then there's no good reason to run the sim")
                print("because you won't know what it did while it ran.")
                exit(1)

    @staticmethod
    def close(signo):
        sys.exit('\n%s received - aborting.' % (str(signo)))

    @staticmethod
    def setup_args(parser):
        parser.add_argument('--lockstep', action='store_true', help='Set if you want sim and FSW to run in lockstep.')
        parser.add_argument('--asap', action='store_true', help='Set if you want sim and FSW to run in lockstep, and as fast as possible (no sleep).')
        parser.add_argument('--loglevel', default="2", type=int, help='Set to how verbose of debug info you want printed. 0=none, 1=more, 2=even more, etc.')
        parser.add_argument('--script', default=None, nargs='+', help="Filename to load for scripted actions, and command-line arguments to give to the script.")
        parser.add_argument('--log', nargs='?', const='', help='The log file type (csv/json/bin) or complete log file name.  Can use strftime formatting for datetime, and a "+" is replaced with "%%Y%%m%%d_%%H%%M%%S"')

    def start_log(self, log_name):
        # hash table of booleans for if each type of message has had it's header
        # put into this log file already.
        self._logged_msg_header = {}
        if log_name.endswith('csv'):
            self.log_file_type = "csv"
        elif log_name.endswith('json'):
            self.log_file_type = "json"
        elif log_name.endswith('bin') or log_name.endswith('log'):
            self.log_file_type = "bin"
        else:
            self.logging.error("ERROR!  Invalid log type " + log_name)
        
        if log_name in ["csv","json","bin","log"]:
            # if they *only* supplied a suffix, assume they want the default datetime format string
            log_file_name = "logs/%Y%m%d_%H%M%S" + "." + self.log_file_type
        else:
            # if they supplied more than just a suffix, then assume they named it the way they want.
            log_file_name = log_name
        log_file_name = log_file_name.replace("+", "%Y%m%d_%H%M%S")
        # if not, generate a filename based on current date/time
        now = datetime.datetime.now()
        log_file_name = now.strftime(log_file_name)
        dir = os. path. dirname(log_file_name)
        if not os.path.exists(dir):
            os.makedirs(dir)
        if self.log_file_type == "bin":
            self.log_file = open(log_file_name, "wb")
        else:
            self.log_file = open(log_file_name, "w")

    def log(self, msg):
        if self.log_file:
            # if the message doesn't have a time set, set it to current sim time.
            if msg.hdr.GetTime() == 0.0:
                msg.hdr.SetTime(self.get_time())

            if self.log_file_type == "json":
                txt = ""
                if not msg.MsgName() in self._logged_msg_header:
                    self._logged_msg_header[msg.MsgName()] = True
                    txt = msg.jsonHeader()
                txt += msg.toJson(includeHeader=True) + "\n"
            elif self.log_file_type == "csv":
                txt = ""
                if not msg.MsgName() in self._logged_msg_header:
                    self._logged_msg_header[msg.MsgName()] = True
                    txt = msg.csvHeader(timeColumn=True)+'\n'
                txt += msg.toCsv(timeColumn=True) + "\n"
            elif self.log_file_type == "bin":
                txt = msg.rawBuffer().raw
            self.log_file.write(txt)

    def process_script_inputs(self):
        while self.script_reader:
            # round time up a smidge to account for floating-point inaccuracies, so
            # that if there's times that print like 4.00000 but are actually 3.999999999,
            # they happen when sim time is 4, and not when it's 4.050
            rounded_time = self.get_time() + self.time_stats.dt/2
            script_line = self.script_reader.get_next_command(rounded_time)
            if script_line == None:
                return

            if script_line.command == "set":
                #TODO: We need to add the ability to override values in a simulation variable database.
                #TODO: These values could be data that the sim is using to construct and send
                #TODO: messages on a regular basis, such that the script can specify values for them,
                #TODO: that are then incorporated into what the sim outputs.
                #TODO: They can also be any other variables that the sim wants to use for any purpose:
                #TODO: some could be just flags to script changing sim behaviour, others could be numerical
                #TODO: values that are nominally inputs to the sim, others could be values used to control
                #TODO: hardware in the lab (ie: power supply voltage, DAC outputs, wind models).
                # It'd be nice to handle JSON here so you could specify the root of a sim variable database branch like:
                # simdb.gps.pos {"lat": 42, "lon": 37, "alt": 12345}
                # instead of having to set the variables separately.
                # after setting values (to JSON for a branch, or just one value for a leaf), we want to be able to
                # restore the ability for the sim to compute values again, with some syntax, maybe this:
                # 0 set simdb.gps.pos UNLOCK
                # ^ That will, for the given branch, iterate over all the leaves under it, and set a flag that lets the sim compute values for them
                pass
            elif script_line.command == "send":
                #TODO: This constructs and sends a message, which is certainly a thing we want the capability to do!
                try:
                    msg = Message.fromJson(script_line.option)
                    # what to do with message?
                    # give it to FSW, or Sim, or both?!?
                    self.logging.warning("Script input %s" % (script_line.lineinfo))
                    self.send(msg, None)
                except json.decoder.JSONDecodeError:
                    self.logging.error("Script ERROR! Ignoring invalid JSON %s" % (script_line.lineinfo))
                    exit(1)
            elif script_line.command == "sleep" or script_line.command == "delay":
                # Nothing to do here, because we already delayed for the sleep/delay, as part of
                # how we handle all script commands
                self.logging.warning("Script %s until %f" % (script_line.command, script_line.exec_time))
            elif script_line.command.startswith("--"):
                # Set command line arguments in the _args structure.
                # Note that this will only change the value, it will not invoke any special handling
                # for the argument that might occur on startup.  If you do want special handling to
                # occur, the code should be restructured so there's a function that gets executed
                # when arguments are set, and that function should be called here as well as on startup.
                command = script_line.command.replace("--","")
                if command in dir(self._args):
                    arg = getattr(self._args, command)
                    if type(arg) == bool:
                        bool_value = not script_line.option.lower() in ['false', '0']
                        self.logging.info("Script %s bool_value: %s = %s" % (command, script_line.option, str(bool_value)))
                        setattr(self._args, command, bool_value)
                    elif type(arg) == float:
                        float_value = float(script_line.option)
                        self.logging.info("Script %s float_value: %s = %s" % (command, script_line.option, str(float_value)))
                        setattr(self._args, command, float_value)
                    elif type(arg) == int:
                        int_value = int(script_line.option)
                        self.logging.info("%s int_value: %s = %s" % (command, script_line.option, str(int_value)))
                        setattr(self._args, command, int_value)
                    else:
                        setattr(self._args, command, script_line.option)
                        print(type(arg))
                        print(dir(arg))
                else:
                    self.logging.error("ERROR! No valid command-line option %s" % command)
                    exit(1)
            elif script_line.command == "exit":
                self.logging.error("Exiting because %s" % (script_line.lineinfo))
                exit(1)
            else:
                # Should add handling of other script commands like:
                # 1) Changing simulation variables to be dynamically calculated vs. set from the script
                #    This would be used so you can temporarily override a value, and then go back to letting the sim calculate it.
                #    Ideally we'd want to be able to set sim variables to functions like a ramp between values over a time period, a sine wave with constants A,B,C for s = A * sin(B*time) + C, etc.
                # 2) Testing conditions of simulation variables being an exact match, within a range, below a min, or above a max.
                #    This would be used to make automated tests decide pass/fail criteria, and log the result to an output file
                # 3) Prompt the user to perform an action
                # 4) Pause or resume running (do we pause physics, execution, or both?)  AviSim would pause physics but not execution, that would horribly confuse an autopilot (integrator windup!)
                # 5) Sleep/Delay script processing for a certain amount of time.  This is in contrast to current behaviour where all lines have a delta T to delay as their first param
                self.logging.error("Unrecognized script command %s" % (script_line.lineinfo))

    def run(self):
        while True:
            self.process_script_inputs()
            self.get_commands()
            
            if self._args.lockstep:
                if not self.time_stats.ready_to_run:
                    # If we're not ready to run, do a gevent sleep to let other
                    # coroutines run.
                    gevent.sleep(0)
                    continue
                self.time_stats.ready_to_run = False
            
            # execute the sim for one time step
            self.fdm.run()
            
            # Send tick to FSW, to allow it to run in sync with sim.
            self.output_time_tick()

            # run flight software, if it exists
            if self.fsw:
                self.fsw.run()

            # send aircraft state
            self.output_aircraft_state()

            # sleep until time to run again
            self.time_stats.time_tick(self.get_time())

    #TODO This is messy, and might be simplified by creating two Client
    # objects, one for FSW and one for FDM.  Client already passes messages
    # between instances of itself, so it might eliminate some of the logic
    # here and in recv().
    def send(self, msg, sender=None):
        # set time, if it's not already set
        if msg.hdr.GetTime() == 0.0:
            msg.hdr.SetTime(self.get_time())
        # log the message
        self.log(msg)

        # give the message to whoever didn't send it

        if sender != self.fdm:
            self.send_to_fdm(msg)

        if self.fsw and sender != self.fsw:
            self.fsw.send(msg.rawBuffer().raw)

        # Also give the message to the network, if it's connected
        if self.cxn:
            try:
                ret = self.cxn.send(msg)
            except (BrokenPipeError, ConnectionResetError):
                self.cxn = None

    def send_to_fdm(self, msg):
        # if it's for setting a sim value, set the value
        #TODO: This should be modified to set things either within the self.fdm[] object, or outside it
        #TODO: and we should be able to override sim values temporarily, and restore them to letting the sim calculate them.
        #TODO: Maybe that means we should require the string to start with "fdm." if it's for the flight dynamics model,
        #TODO: and otherwise make it be treated like a "normal" simulation variable (which doesn't exist yet)
        if type(msg) == Messaging.Messages.Hitl.SetSimValue:
            print("  ---- Setting %s = %f" % (msg.GetVariableName(), msg.Value))
            self.fdm[msg.GetVariableName()] = msg.Value
        
    def recv(self):
        # if there's any messages from flight software, return them,
        # but also send them to the network
        if self.fsw:
            msgbuf = self.fsw.recv()
            if msgbuf:
                hdr = Messaging.hdr(messageBuffer=msgbuf)
                msg = Messaging.MsgFactory(hdr)
                # when logging, always override received message time with sim
                # time, so all messages have a consistent timestamp.
                msg.hdr.SetTime(self.get_time())
                self.log(msg)
                if self.cxn:
                    try:
                        self.cxn.send(msg)
                    except (BrokenPipeError, ConnectionResetError):
                        self.cxn = None
                return msg
        # if there's any messages from the network, return them,
        # but also send them to flight software and FDM
        if self.cxn:
            try:
                msg = self.cxn.recv()
                if msg:
                    # when logging, always override received message time with sim
                    # time, so all messages have a consistent timestamp.
                    msg.hdr.SetTime(self.get_time())
                    self.log(msg)
                    if self.fsw:
                        self.fsw.send(msg.rawBuffer().raw)
                    self.send_to_fdm(msg)
                    return msg
            except (BrokenPipeError, ConnectionResetError):
                self.cxn = None
        # since nothing returned above, return None to indicate no messages.
        return None

    def process_sim_command(self, msg):
        if type(msg) == Messaging.Messages.TimeTock:
            self.time_stats.ready_to_run = True

    def get_commands(self):
        while True:
            msg = self.recv()
            if msg == None:
                self.logging.debug("No messages read, are you sure FSW is running?")
                break
            self.process_sim_command(msg)

    def output_time_tick(self):
        if self._args.lockstep or self._args.fsw:
            simtime = self.get_time()
            tick = Messaging.Messages.TimeTick()
            tick.SetTime(simtime)
            self.send(tick, self.fdm)



#!/usr/bin/env python3
import sys
import os
import string
from time import gmtime, strftime

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.parser.MsgUtils import *

def Usage():
    sys.stderr.write('Usage: ' + sys.argv[0] + ' msgdir outputfile\n')
    sys.exit(1)

def Messages(inputData):
    return inputData["Messages"]

def ProcessDir(outFile, msgDir, subdirComponent, isHeaderDir):
    for filename in sorted(os.listdir(msgDir)):
        global inputFilename
        inputFilename = msgDir + '/' + filename
        if os.path.isdir(inputFilename):
            subdirParam = filename
            if subdirComponent != "":
                subdirParam = subdirComponent + "/" + subdirParam
            ProcessDir(outFile, inputFilename, subdirParam, filename=='headers' or isHeaderDir)
        else:
            try:
                inputData, inputFileHash = readFile(inputFilename)
                if inputData != 0:
                    ProcessFile(filename, outFile, inputData, subdirComponent, isHeaderDir)
            except MessageException as e:
                sys.stderr.write('Error in ' + inputFilename)
                sys.stderr.write('\n'+str(e)+'\n')
                outFile.close()
                os.remove(outFile.name)
                sys.exit(1)
            except:
                sys.stderr.write("\nError in " + inputFilename + "!\n\n")
                outFile.close()
                os.remove(outFile.name)
                raise

def fieldTypeValid(field):
    allowedFieldTypes = \
    ["uint64", "uint32", "uint16", "uint8",
      "int64",  "int32",  "int16",  "int8",
      "float64", "float32"]
    return field["Type"] in allowedFieldTypes

def ProcessMsg(filename, msg, subdirComponent, enums, isHeader):
    enumNames = {}
    for enum in enums:
        enumNames[enum["Name"]] = enum
    
    if not isHeader:
        id = msgID(msg, enums, -1)
        idInt = int(id, 0)
        if idInt:
            global msgNames
            fullMsgName = subdirComponent+'/'+msgName(msg)
            if idInt in msgNames:
                raise MessageException('\nERROR! '+fullMsgName+' uses id '+str(id)+', but already used by '+msgNames[idInt]+'\n\n')
            if fullMsgName in msgNames:
                raise MessageException('\nERROR! '+fullMsgName+' being processed, but name already used by '+msg['Name']+'.\n\n')
            msgNames[idInt] = fullMsgName
    if "Fields" in msg:
        fieldNames = {}
        for field in msg["Fields"]:
            bitOffset = 0
            if not re.match("^[\w\d_]+$", field["Name"]):
                raise MessageException('bad field name [' + field["Name"]+"] in message "+msgName(msg))
            if field["Name"] in fieldNames:
                raise MessageException('Duplicate field name [' + field["Name"]+"] in message "+msgName(msg))
            fieldNames[field["Name"]] = field["Name"]
            if not fieldTypeValid(field):
                raise MessageException('field ' + field["Name"] + ' has invalid type ' + field["Type"]+ " in message "+msgName(msg))
            if "Enum" in field:
                if not field["Enum"] in enumNames:
                    raise MessageException('bad enum [' + field["Enum"]+"] in message "+msgName(msg))
                if " " in field["Enum"]:
                    raise MessageException('bad enum [' + field["Enum"]+"] in message "+msgName(msg))
                pass
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    bitOffset += numBits
                    if not re.match("^[\w\d_]+$", bits["Name"]):
                        raise MessageException('bad bitfield name [' + bits["Name"]+"] in message "+msgName(msg))
                    if bits["Name"] in fieldNames:
                        raise MessageException('Duplicate bitfield name [' + bits["Name"]+"] in field ["+field["Name"]+"] in message "+msgName(msg))
                    fieldNames[bits["Name"]] = bits["Name"]
                if bitOffset > 8*fieldSize(field):
                    raise MessageException('too many bits in message '+msgName(msg))
                if "Enum" in bits:
                    if not bits["Enum"] in enumNames:
                        raise MessageException('bad enum in message '+msgName(msg))
                    pass
    if isHeader:
        return ''
    return (subdirComponent+'/'+msgName(msg)).ljust(35) +" "+(subdirComponent+'/'+filename).ljust(40) +" "+ str(id).rjust(10)+'\n'

def ProcessFile(filename, outFile, inputData, subdirComponent, isHeader):
    enums = Enums(inputData)
    ids = MsgIDs(inputData)
    
    PatchStructs(inputData)

    if "Messages" in inputData:
        for msg in Messages(inputData):
            msg["ids"] = ids
            outFile.write(ProcessMsg(filename, msg, subdirComponent, enums, isHeader))

def main(args=None):
    if len(sys.argv) < 3:
        Usage();
    outputFilename = sys.argv[1]
    msgDir = sys.argv[2]
    
    global msgNames
    msgNames = {}

    try:
        os.makedirs(os.path.dirname(outputFilename))
    except:
        pass
    with open(outputFilename,'w') as outFile:
        # loop over input message files
        ProcessDir(outFile, msgDir, "", msgDir.endswith('/headers'))
        if len(sys.argv) > 3:
            msgDir = sys.argv[3]
            ProcessDir(outFile, msgDir, "", msgDir.endswith('/headers'))

# main starts here
if __name__ == '__main__':
    main()



import sys
import subprocess
import os

from PyQt5 import QtCore, QtGui, QtWidgets

base_dir = "/opt/msgtools"
msg_dir = base_dir + "/messages"
obj_dir = base_dir + "/obj/CodeGenerator"

class CodegenGui(QtCore.QObject):
    languageOptions = {}
    languageOptions['C']          = ["c", "C"]
    languageOptions['C++']        = ["cpp", "Cpp"]
    languageOptions['Dart']       = ["dart", "Dart/lib"]
    languageOptions['Java']       = ["java", "Java"]
    languageOptions['Javascript'] = ["javascript", "Javascript"]
    languageOptions['Matlab']     = ["matlab", "Matlab/+Messages"]
    languageOptions['Python']     = ["python", "Python"]
    languageOptions['Html']       = ["HTML", "language.py", "Template.html", "HeaderTemplate.html"]
    "find $(MSGDIR)/Html -type d -print0 | xargs -n 1 -0 cp $(CG_DIR)HTML/bootstrap.min.css"
    def __init__(self, menubar, settings):
        QtCore.QObject.__init__(self)
        self.settings = settings
        codegenMenu = menubar.addMenu('&Generate Code')

        cleanAction = QtWidgets.QAction('&Clean', self)
        cleanAction.triggered.connect(self.codegen_clean)
        codegenMenu.addAction(cleanAction)

        buildAction = QtWidgets.QAction('&Build', self)
        buildAction.triggered.connect(self.codegen)
        codegenMenu.addAction(buildAction)
        
        submenu = codegenMenu.addMenu('Languages')
        languages = self.language_options()
        allAction = QtWidgets.QAction('All', self)
        allAction.triggered.connect(self.language_selected)
        allAction.setCheckable(True)
        submenu.addAction(allAction)
        self.languageCheckboxes = []
        self.selectedLanguages = self.settings.value("languages", "").split(',')
        # remove empty string if it's there
        try:
            self.selectedLanguages.remove('')
        except ValueError:
            pass
        for language in languages:
            lAction = QtWidgets.QAction(language, self)
            lAction.triggered.connect(self.language_selected)
            lAction.setCheckable(True)
            submenu.addAction(lAction)
            self.languageCheckboxes.append(lAction)
            if language in self.selectedLanguages:
                lAction.setChecked(True)

    def language_selected(self):
        s = self.sender()
        checked = s.isChecked()
        language_name = s.text()
        if language_name == 'All':
            for languageCheckbox in self.languageCheckboxes:
                languageCheckbox.setChecked(checked)
        else:
            if checked:
                if not language_name in self.selectedLanguages:
                    self.selectedLanguages.append(language_name)
            else:
                if not language_name in self.selectedLanguages:
                    self.selectedLanguages.remove(language_name)

    def save_settings(self):
        self.settings.setValue("languages", ','.join(self.selectedLanguages))

    def codegen(self):
        self.build(self.selectedLanguages)

    def codegen_clean(self):
        self.clean()
        
    def build(self, languageNames):
        commands = []
        for languageName in languageNames:
            languageOptions = self.languageOptions[languageName]
            languagePluginName = languageOptions[0]
            outputDir = obj_dir + '/' + languageOptions[1]
            invoke = 'msgparser %s %s %s' % (msg_dir, outputDir, languagePluginName)
            commands.append(invoke)
        self.invoke(commands)

    def clean(self):
        self.invoke(['rm -rf %s' % obj_dir])

    def language_options(self):
        return self.languageOptions.keys()

    def invoke(self, commands):
        dlg = CodeGeneratorDialog(commands)
        code = dlg.exec_()

class CodeGeneratorDialog(QtWidgets.QDialog):
    def __init__(self, commands, parent = None):
        QtWidgets.QDialog.__init__(self, parent)
        self.resize(600, 400);

        vlayout = QtWidgets.QVBoxLayout()
        self.setLayout(vlayout)

        self.textBox = QtWidgets.QPlainTextEdit()
        self.textBox.setReadOnly(True)
        vlayout.addWidget(self.textBox)

        self.ok = QtWidgets.QPushButton("OK")
        self.ok.setEnabled(False)
        self.ok.clicked.connect(self.accept)
        vlayout.addWidget(self.ok)

        self.commands = commands
        self.start_process(self.commands.pop())
    
    def start_process(self, command):
        self.textBox.insertPlainText('> %s\n' % command)
        self.textBox.moveCursor(QtGui.QTextCursor.End)
        self.proc = QtCore.QProcess(self)
        self.proc.finished.connect(self.process_exited)
        self.proc.readyReadStandardOutput.connect(self.print_output)
        self.proc.readyReadStandardError.connect(self.print_output)
        self.proc.setWorkingDirectory(base_dir)
        self.proc.start(command)

    def process_exited(self):
        self.textBox.moveCursor(QtGui.QTextCursor.End)
        self.textBox.insertPlainText("\nDone!\n")
        self.textBox.moveCursor(QtGui.QTextCursor.End)
        if len(self.commands) > 0:
            self.start_process(self.commands.pop())
        else:
            #self.accept()
            self.ok.setEnabled(True)

    def print_output(self):
        QtCore.QCoreApplication.processEvents()
        self.textBox.moveCursor(QtGui.QTextCursor.End)
        self.textBox.insertPlainText(self.proc.readAllStandardOutput().data().decode('utf-8'))
        self.textBox.insertPlainText(self.proc.readAllStandardError().data().decode('utf-8'))
        self.textBox.moveCursor(QtGui.QTextCursor.End)



#!/usr/bin/en"v python3
from PyQt5 import QtCore, QtGui, QtWidgets
import can
import ctypes

from msgtools.lib.messaging import Messaging, FieldInfo, offset, size
from msgtools.lib.header_translator import HeaderTranslator
from msgtools.server import CanPortDialog
import struct
import sys
import time

def hexbytes(d):
    bytes_per_line = 8
    ret = "%d\n" % len(d)
    bytes = 0
    while bytes < len(d):
        ret += ":".join("{:02X}".format(c) for c in d[bytes:bytes+bytes_per_line]) + "\n"
        bytes += bytes_per_line
    return ret

def Crc16(data):
    crc = 0;
    for i in range(0,len(data)):
        d = struct.unpack_from('B', data, i)[0]
        crc = (crc >> 8) | (crc << 8)
        crc ^= d
        crc ^= (crc & 0xff) >> 4
        crc ^= crc << 12
        crc = 0xFFFF & crc
        crc ^= (crc & 0xff) << 5
        crc = 0xFFFF & crc
    return crc

class CanFragmentation(QtCore.QObject):
    statusUpdate = QtCore.pyqtSignal(str)

    MAX_MSG_SIZE = 64
    MAX_PACKET_SIZE = MAX_MSG_SIZE
    MAX_FRAG_PACKET_SIZE = MAX_MSG_SIZE
    MAX_FRAG_FIRST_PACKET_SIZE = MAX_MSG_SIZE

    # DLC length coding
    DLC_LENGTH_CODING = [
        0, 1, 2, 3, 4, 5, 6, 7, 8,
        12, # 9
        16, # 10
        20, # 11
        24, # 12
        32, # 13
        48, # 14
        64  # 15
    ]

    DLC_FROM_LENGTH = [
        0,1,2,3,4,5,6,7,8,
        9,9,9,9, # 9-12
        10,10,10,10, # 13-16
        11,11,11,11, # 17-20
        12,12,12,12, # 21-24
        13,13,13,13,13,13,13,13, # 25-32
        14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14, # 33-48
        15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15  # 49-64
    ]
    
    @staticmethod
    def length_decrement(packet_len):
        #print("%d / %d" % (packet_len, len()
        return CanFragmentation.DLC_LENGTH_CODING[CanFragmentation.DLC_FROM_LENGTH[packet_len]] - packet_len

    @staticmethod
    def packet_count(msg_len):
        if msg_len <= CanFragmentation.MAX_PACKET_SIZE:
            return 1
        total_bytes_for_frag_packets = msg_len+(CanFragmentation.MAX_FRAG_PACKET_SIZE-CanFragmentation.MAX_FRAG_FIRST_PACKET_SIZE)
        return int(total_bytes_for_frag_packets / CanFragmentation.MAX_FRAG_PACKET_SIZE) + (total_bytes_for_frag_packets % CanFragmentation.MAX_FRAG_PACKET_SIZE > 0)

    def __init__(self):
        super(CanFragmentation, self).__init__(None)

        import CANHeader
        # Make a subclass of CANHeader that acts like it has a "DataLength".
        # The real CAN header doesn't have a DataLength, it instead has a "DLC"
        # (which for CAN FD is larger than the actual length of the message body),
        # and a LengthDecrement field which is how much smaller the real length
        # is than is indicated by DLC.
        class CANHeaderWithLength(CANHeader.CANHeader):
            def __init__(self, messageBuffer=None):
                super(CANHeaderWithLength, self).__init__(messageBuffer)
                self.length = -1
            @offset('-1')
            @size('0')
            def GetDataLength(self):
                #print("  CANHeaderWithLength.GetDataLength() = %d" % self.length)
                return self.length
            def SetDataLength(self, l):
                self.length = l
                #print("  CANHeaderWithLength.SetDataLength(%d)" % self.length)
        CANHeaderWithLength.fields.append(
            FieldInfo(name="DataLength",type="int",units="",minVal="0",maxVal="65535",description="Length of CAN Message.",get=CANHeaderWithLength.GetDataLength,set=CANHeaderWithLength.SetDataLength,count=1, bitfieldInfo = [], enum = [])
        )
        
        # If fragmentation headers exist, use them to perform fragmentation and reassembly of CAN messages.
        try:
            self.frag_hdr = CANHeader.FragmentationHeader
            self.frag_hdr_start = CANHeader.FragmentationHeaderStart

        except Messaging.LoadError:
            self.frag_hdr = None
            self.frag_hdr_start = None
        if self.frag_hdr and self.frag_hdr_start:
            # find bytes in most fragmented packets by taking max size and subtracting size of fragmentation header
            CanFragmentation.MAX_FRAG_PACKET_SIZE = CanFragmentation.MAX_PACKET_SIZE - self.frag_hdr.SIZE
            # find bytes in first fragmented packet by also subtracting size of the fragmentation 'start' header
            CanFragmentation.MAX_FRAG_FIRST_PACKET_SIZE = CanFragmentation.MAX_FRAG_PACKET_SIZE - self.frag_hdr_start.SIZE
            # total size is size of first frag packet, plus all the other fragmented packets.
            max_packets = int(self.frag_hdr.GetSequenceNumber.maxVal)
            CanFragmentation.MAX_MSG_SIZE = CanFragmentation.MAX_FRAG_FIRST_PACKET_SIZE + (max_packets-1)* CanFragmentation.MAX_FRAG_PACKET_SIZE

        self.hdr = CANHeaderWithLength

        self.hdrTranslator = HeaderTranslator(self.hdr, Messaging.hdr)

        self.frag_stream_id = 0
        
        self.use_print = False
        
        self.in_progress_reassembly = {}

    def print(self, s):
        if self.use_print:
            print("F: " + s)
        else:
            self.statusUpdate.emit("Fragmentation: "+s)

    def fragment(self, networkMsg):
        # translate the header
        canHdr = self.hdrTranslator.translateHdr(networkMsg)
        # if we can't translate the header, just return because the message can't be sent on CAN
        if canHdr == None:
            if Messaging.debug:
                self.print("Can't translate 0x%x to CAN message" % networkMsg.GetMessageID())
            return []

        # check that the message fits
        if networkMsg.GetDataLength() > CanFragmentation.MAX_MSG_SIZE:
            if Messaging.debug:
                self.print("Message 0x%x size %d > %d bytes, too big for CAN" % (networkMsg.GetMessageID(), networkMsg.GetDataLength(), CanFragmentation.MAX_MSG_SIZE))
            return []

        packets = []
        
        bytes_sent = 0
        packets_sent = 0
        msg_len = networkMsg.GetDataLength()
        max_packets = CanFragmentation.packet_count(msg_len)
        #self.print("\nmsg_len: %d\nmax_packets: %d\n" % (msg_len, max_packets))
        data_ptr = networkMsg.rawBuffer()[type(networkMsg).SIZE:]
        while bytes_sent < msg_len:
            if max_packets == 1:
                max_packet_size = CanFragmentation.MAX_PACKET_SIZE
            else:
                max_packet_size = CanFragmentation.MAX_FRAG_PACKET_SIZE if bytes_sent else CanFragmentation.MAX_FRAG_FIRST_PACKET_SIZE
            packet_size = min(msg_len - bytes_sent, max_packet_size)
            #self.print("packet_size: %d, max_packet_size: %d" % (packet_size, max_packet_size))
            if max_packets > 1:
                canHdr.SetFragmented(1)
                frag_hdr = self.frag_hdr()
                frag_hdr.SetNewStream(1 if bytes_sent == 0 else 0)
                frag_hdr.SetStreamID(self.frag_stream_id)
                self.frag_stream_id += 1
                if self.frag_stream_id > int(frag_hdr.SetStreamID.maxVal):
                    self.frag_stream_id = 0
                frag_hdr.SetSequenceNumber(max_packets if bytes_sent == 0 else packets_sent)
                tx_data = frag_hdr.rawBuffer().raw
                if bytes_sent == 0:
                    frag_hdr_start = self.frag_hdr_start()
                    frag_hdr_start.SetMessageLength(msg_len)
                    frag_hdr_start.SetCRC16(Crc16(data_ptr))
                    tx_data += frag_hdr_start.rawBuffer().raw
                tx_data += data_ptr[bytes_sent:bytes_sent+packet_size]
            else:
                tx_data = data_ptr[0:packet_size]
            #self.print(hexbytes(tx_data))
            unpadded_len = len(tx_data)
            tx_data = tx_data + ctypes.create_string_buffer(CanFragmentation.length_decrement(len(tx_data)))
            tx_frame = can.Message(
                is_extended_id=True,
                is_remote_frame=False,
                is_error_frame=False,
                bitrate_switch=True,
                is_fd=True,
                data=tx_data)
            # compute the LengthDecrement after DLC is set in header based on body length,
            # and then update arbitration ID which contains LengthDecrement
            canHdr.SetLengthDecrement(CanFragmentation.length_decrement(unpadded_len))
            tx_frame.arbitration_id = canHdr.GetCanID()
            packets.append(tx_frame)
            #print("tx " + str(tx_frame))
            #print("sent %d/%d/%d of %d/%d (%d)" % (packet_size, unpadded_len, len(tx_data), bytes_sent, networkMsg.GetDataLength(), canHdr.GetLengthDecrement()))
            bytes_sent += packet_size
            packets_sent += 1
        return packets
    
    def reassemble(self, rx_frame):
        class ReassemblyProgress:
            def __init__(self, networkMsg, rx_byte_count, last_time):
                self.networkMsg = networkMsg
                self.rx_byte_count = rx_byte_count
                self.last_time = last_time
        
        #TODO: Need to look through in-progress table for stale partially assembled
        # messages and delete them (and probably print error messages too)
        
        # consider enforcing rules for how many in-progress messages a single source can
        # have, and throw away that source's in progress messages if too many come in with
        # unique stream IDs

        canHdr = self.hdr()
        canHdr.SetCanID(rx_frame.arbitration_id)
        canHdr.SetDataLength(len(rx_frame.data) - canHdr.GetLengthDecrement())
        #print(canHdr)
        #print("in[%d]: %s" % (canHdr.GetDataLength(), str(canHdr)))
        #print("rx " + str(rx_frame))
        if self.frag_hdr and canHdr.GetFragmented():
            frag_hdr = self.frag_hdr(rx_frame.data)
            key = "%s@%s" % (str(canHdr.GetID()), "/".join(Messaging.HeaderRoute(canHdr)))
            #print(frag_hdr)
            if frag_hdr.GetNewStream():
                packet_count = frag_hdr.GetSequenceNumber()
                frag_hdr_start = self.frag_hdr_start(rx_frame.data[self.frag_hdr.SIZE:])
                
                packet_data = rx_frame.data[self.frag_hdr.SIZE+self.frag_hdr_start.SIZE:canHdr.GetDataLength()]
                #print("Adding %d/%d/%d to %d/%d (%d)" % (len(packet_data), canHdr.GetDataLength(), len(rx_frame.data), 0, frag_hdr_start.GetMessageLength(), canHdr.GetLengthDecrement()))
                #print(frag_hdr_start)
                #print("Making %d byte body with %d bytes to start" % (frag_hdr_start.GetMessageLength(), len(packet_data)))
                canHdr.SetDataLength(frag_hdr_start.GetMessageLength())
                #print("->")
                networkMsg = self.hdrTranslator.translateHdr(canHdr)
                #print("<-")
                if networkMsg:
                    #print(networkMsg)
                    # copy packet body into message buffer
                    rx_byte_count=0
                    for b in packet_data:
                        networkMsg.rawBuffer()[type(networkMsg).SIZE+rx_byte_count] = b
                        rx_byte_count += 1
                    self.in_progress_reassembly[key] = ReassemblyProgress(networkMsg,rx_byte_count,time.time())
                    #print("Stored %d bytes for network msg %s" % (rx_byte_count, str(networkMsg)))
            else:
                packet_data = rx_frame.data[self.frag_hdr.SIZE:canHdr.GetDataLength()]
                if key in self.in_progress_reassembly:
                    progress = self.in_progress_reassembly[key]
                    #print("Adding %d/%d/%d to %d/%d (%d)" % (len(packet_data), canHdr.GetDataLength(), len(rx_frame.data), progress.rx_byte_count, progress.networkMsg.GetDataLength(), canHdr.GetLengthDecrement()))
                    for b in packet_data:
                        if progress.rx_byte_count > progress.networkMsg.GetDataLength():
                            self.print("ERROR!  On byte %d/%d but only have room for %d" % (progress.rx_byte_count, len(packet_data), progress.networkMsg.GetDataLength()))
                        progress.networkMsg.rawBuffer()[type(progress.networkMsg).SIZE+progress.rx_byte_count] = b
                        progress.rx_byte_count += 1
                    #print("Found network header " + str(progress.networkMsg))
                    #print("Got to %d/%d bytes!" % (progress.rx_byte_count, progress.networkMsg.GetDataLength()))
                    if progress.rx_byte_count >= progress.networkMsg.GetDataLength():
                        #print("Reassembled a msg!!")
                        self.in_progress_reassembly.pop(key)
                        return progress.networkMsg
                    #else:
                    #    print("Have %d/%d" % (progress.rx_byte_count, progress.networkMsg.GetDataLength()))
        else:
            #self.print("reassembling " + str(rx_frame) + hexbytes(rx_frame.data))
            #self.print("time is " + str(rx_frame.timestamp))
            networkMsg = self.hdrTranslator.translateHdrAndBody(canHdr, rx_frame.data)
            try:
                networkMsg.SetTime(rx_frame.timestamp)
            except struct.error:
                networkMsg.SetTime(int(rx_frame.timestamp))
            return networkMsg
        return None
    
class CanConnection(QtCore.QObject):
    statusUpdate = QtCore.pyqtSignal(str)
    messagereceived = QtCore.pyqtSignal(object)
    disconnected = QtCore.pyqtSignal(object)

    def __init__(self):
        super(CanConnection, self).__init__(None)

        self.fragmentation = CanFragmentation()
        self.fragmentation.statusUpdate.connect(self.print)
        
        self.base_name = "CAN"
        self.settings = QtCore.QSettings("MsgTools", "MessageServer/"+self.base_name)
        
        self.removeClient = QtWidgets.QPushButton("Remove")
        self.removeClient.pressed.connect(self.onDisconnected)

        # button to open/close CAN port
        self.openCloseButton = QtWidgets.QPushButton("button")
        self.openCloseButton.pressed.connect(self.openCloseSwitch)
        
        # button select new CAN port
        self.selectPortButton = QtWidgets.QPushButton("Select Port")
        self.selectPortButton.pressed.connect(self.selectPort)

        self.statusLabel = QtWidgets.QLabel()
        self.rxMsgCount = 0
        self.subscriptions = {}
        self.subMask = 0
        self.subValue = 0
        self.isHardwareLink = True
        
        # default to last used port
        self.portName = self.settings.value("portName", None)
        self.interfaceName = self.settings.value("interfaceName", "socketcan")

        self.canPort = None
        self.socketReadNotifier = None
        self.socketErrorNotifier = None
        self.name = "%s.%s" % (self.interfaceName, self.portName)

        self.statusLabel.setText(self.name)
        
        self.reopenTimer = QtCore.QTimer(self)
        self.reopenTimer.setInterval(1000)
        self.reopenTimer.timeout.connect(self._openCanPort)

    def print(self, s):
        self.statusUpdate.emit("CAN: "+s)

    def onDisconnected(self):
        self.disconnected.emit(self)
    
    def onErrorOccurred(self, error):
        if error == QCanBusDevice.NoError:
            pass
        elif error == QCanBusDevice.ReadError:
            self.print("Error, Closing "+str(self.name))
            self._closeCanPort()
            # Don't restart the timer if it's active.
            # If we did, then a steady stream of errors would keep resetting
            # the timer, and it would never go off.
            if not self.reopenTimer.isActive():
                self.reopenTimer.start()
        #elif error == QCanBusDevice.: # are there any errors we should just ignore?
        #    pass

    def widget(self, index):
        if index == 0:
            return self.removeClient
        if index == 1:
            return self.openCloseButton
        if index == 2:
            return self.selectPortButton
        if index == 3:
            return self.statusLabel
        return None
    
    def _openCanPort(self):
        if self.canPort:
            # if we're already open, make sure we stop retrying to open
            self.reopenTimer.stop()
        else:
            try:
                # open the CAN bus
                self.canPort = can.Bus(channel=self.portName, interface=self.interfaceName, fd=True)
                # create a socket notifier, so we get a signal when data is ready
                self.socketReadNotifier = QtCore.QSocketNotifier(self.canPort.socket.fileno(), QtCore.QSocketNotifier.Read)
                self.socketReadNotifier.activated.connect(self.onReadReady)
                self.socketErrorNotifier = QtCore.QSocketNotifier(self.canPort.socket.fileno(), QtCore.QSocketNotifier.Exception)
                self.socketErrorNotifier.activated.connect(self.onReadError)
                
                self.print("Opened "+self.name)
                self.openCloseButton.setText("Close")
                self.settings.setValue("portName", self.portName)
                self.settings.setValue("interfaceName", self.interfaceName)
                self.reopenTimer.stop()
                return True
            except Exception as ex:
                if type(ex) == ValueError:
                    self.print("Parameters are out of range")
                elif type(ex) == can.CanInterfaceNotImplementedError:
                    self.print("The driver cannot be accessed")
                elif type(ex) == can.CanInitializationError:
                    self.print("The bus cannot be initialized")
                else:
                    self.print(str(ex))
                self.canPort = None
                self._closeCanPort()
                return False

    def _closeCanPort(self):
        if self.canPort:
            self.canPort.shutdown()
            self.canPort = None
            self.socketReadNotifier = None
            self.socketErrorNotifier = None
        self.openCloseButton.setText("Open")

    def openCloseSwitch(self):
        # open or close the port
        if self.canPort:
            self._closeCanPort()
            self.print("Closed "+self.name)
        else:
            if not self._openCanPort():
                self.print("Can't open "+self.name+"!")

    def portChanged(self, interface, channel):
        self.interfaceName = interface
        self.portName = channel
        self.name = "%s.%s" % (self.interfaceName, self.portName)
        if self.canPort:
            self._closeCanPort()
        self.statusLabel.setText(self.name)
        self.openCloseSwitch()

    def selectPort(self):
        d = CanPortDialog.SelectCanPortDialog()
        d.portChanged.connect(self.portChanged)
        d.statusUpdate.connect(self.statusUpdate)
        d.exec_()

    def start(self):
        self.openCloseSwitch()

    # need to connect something to this
    def onReadError(self, errType):
        self.print("RX " + errType)

    def stop(self):
        pass

    def onReadReady(self):
        while self.canPort:
            # Read data with zero timeout
            rx_frame = self.canPort.recv(timeout=0.0)
            # if no data, break
            if rx_frame == None:
                break
            networkMessage = self.fragmentation.reassemble(rx_frame)
            if networkMessage:
                self.messagereceived.emit(networkMessage)

    def sendMsg(self, networkMsg):
        if self.canPort == None:
            return
        
        can_packets = self.fragmentation.fragment(networkMsg)
        for tx_frame in can_packets:
            self.canPort.send(tx_frame)

def PluginConnection(param=None):
    return CanConnection()

def PluginEnabled():
    try:
        from CANHeader import CANHeader
        return True
    except:
        return False


import collections
PluginInfo = collections.namedtuple('PluginInfo', ['name', 'enabled', 'connect_function'])
plugin_info = PluginInfo('CAN', PluginEnabled, PluginConnection)



import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

def outputSubdir(outDir, filename):
    return outDir + "/+" + filename

def Mask(numBits):
    return str(2 ** numBits - 1)

def fieldType(field):
    typeStr = field["Type"]
    if "int" in typeStr:
        return typeStr
    if typeStr == "float32":
        return "single";
    if typeStr == "float64":
        return "double";
    return "?"

def fnHdr(field):
    ret = "%% %s %s, (%s to %s)" % (MsgParser.fieldDescription(field), MsgParser.fieldUnits(field), MsgParser.fieldMin(field), MsgParser.fieldMax(field))
    return ret

def typeForScaledInt(field):
    numBits = MsgParser.fieldNumBits(field)
    if numBits > 24:
        return "double"
    return "single"

def matlabFieldName(msg, field):
    fieldName = field["Name"]
    if fieldName == msgShortName(msg):
        fieldName = fieldName + "_"
    return fieldName

def getFn(msg, field):
    loc = str(1+MsgParser.fieldLocation(field))
    end_loc = str(1+MsgParser.fieldLocation(field) + MsgParser.fieldSize(field)*MsgParser.fieldCount(field)-1)
    param = "obj"
    access = "typecast(obj.m_data(%s:%s), '%s')" % (loc, end_loc, fieldType(field))
    if MsgParser.big_endian:
        access = "swapbytes(%s)" % access
    access = getMath(access, field, typeForScaledInt(field))
    retType = fieldType(field)
    if "Offset" in field or "Scale" in field:
        retType = typeForScaledInt(field)
    asInt = ""
    if "Enum" in field:
        asInt = "AsInt"
    ret = '''\
%s
function ret = get.%s%s(%s)
    ret = %s;
end
''' % (fnHdr(field), matlabFieldName(msg,field), asInt, param, access)
    if "Enum" in field:
        ret += '''\
%s
function ret = get.%s(%s)
    ret = obj.%sAsInt;
''' % (fnHdr(field), matlabFieldName(msg,field), param, matlabFieldName(msg,field))
        ret += "    if isKey(obj."+field["Enum"]+"Enum, ret)\n"
        ret += "        ret = obj."+field["Enum"]+"Enum(ret);\n"
        ret += "    end\n"
        ret += "end\n"
    # Need to handle arrays of structs, where elements of array are not contiguous!
    if MsgParser.fieldSize(field) != MsgParser.fieldArrayElementOffset(field):
        ret = "%%ERROR! Need to handle arrays of structs, where elements of array are not contiguous!\n" + ret 
    return ret

def setFn(msg, field):
    valueString = setMath("value", field, fieldType(field))
    loc = str(1+MsgParser.fieldLocation(field))
    end_loc = str(1+MsgParser.fieldLocation(field) + MsgParser.fieldSize(field)*MsgParser.fieldCount(field)-1)
    asInt = ""
    if "Enum" in field:
        asInt = "AsInt"
    ret = '''\
%s
function obj = set.%s%s(obj, value)
''' % (fnHdr(field), matlabFieldName(msg,field), asInt)
    if MsgParser.big_endian:
        swap = "swapbytes"
    else:
        swap = ""
    ret += '''\
    obj.m_data(%s:%s) = typecast(%s(%s(%s)), 'uint8');
end
''' % (loc, end_loc, swap, fieldType(field), valueString)
    if "Enum" in field:
        ret += '''\
%s
function obj = set.%s(obj, value)
''' % (fnHdr(field), matlabFieldName(msg,field))
        ret += "    if isKey(obj.Reverse"+field["Enum"]+"Enum, value)\n"
        ret += "        value = obj.Reverse"+field["Enum"]+"Enum(value);\n"
        ret += "    end\n"
        ret += "    obj."+matlabFieldName(msg,field)+"AsInt = value;\n"
        ret += "end\n"
    # Need to handle arrays of structs, where elements of array are not contiguous!
    if MsgParser.fieldSize(field) != MsgParser.fieldArrayElementOffset(field):
        ret = "%%ERROR! Need to handle arrays of structs, where elements of array are not contiguous!\n" + ret 
    return ret

def getBitsFn(msg, field, bits, bitOffset, numBits):
    access = "bitand(bitshift(obj.%s, -%s), %s)" % (matlabFieldName(msg,field), str(bitOffset), Mask(numBits))
    access = getMath(access, bits, typeForScaledInt(bits))
    retType = fieldType(field)
    if "Offset" in bits or "Scale" in bits:
        retType = typeForScaledInt(bits)
    asInt = ""
    if "Enum" in bits:
        asInt = "AsInt"
    ret = '''\
%s
function ret = get.%s%s(obj)
    ret = %s;
end
''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), asInt, access)
    if "Enum" in bits:
        ret += '''\
%s
function ret = get.%s(obj)
    ret = obj.%sAsInt;
''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), MsgParser.BitfieldName(field, bits))
        ret += "    if isKey(obj."+bits["Enum"]+"Enum, ret)\n"
        ret += "        ret = obj."+bits["Enum"]+"Enum(ret);\n"
        ret += "    end\n"
        ret += "end\n"
    return ret

def setBitsFn(msg, field, bits, bitOffset, numBits):
    paramType = fieldType(field)
    valueString = setMath("value", bits, fieldType(field))
    if "Offset" in bits or "Scale" in bits:
        paramType = typeForScaledInt(bits)
    asInt = ""
    if "Enum" in bits:
        asInt = "AsInt"
    ret = '''\
%s
function obj = set.%s%s(obj, value)
''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), asInt)
    ret += '''\
    obj.%s = bitor(bitand(obj.%s, bitcmp(bitshift(%s(%s),%s))), (bitshift((bitand(%s, %s)), %s)));
end''' % (matlabFieldName(msg,field), matlabFieldName(msg,field), fieldType(field), Mask(numBits), str(bitOffset), valueString, Mask(numBits), str(bitOffset))
    if "Enum" in bits:
        ret += '''\
%s
function obj = set.%s(obj, value)
''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits))
        ret += "    if isKey(obj.Reverse"+bits["Enum"]+"Enum, value)\n"
        ret += "        value = obj.Reverse"+bits["Enum"]+"Enum(value);\n"
        ret += "    end\n"
        ret += "    obj."+MsgParser.BitfieldName(field, bits)+"AsInt = value;\n"
        ret += "end\n"
    return ret

def accessors(msg):
    gets = []
    sets = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            gets.append(getFn(msg,field))
            sets.append(setFn(msg,field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    gets.append(getBitsFn(msg, field, bits, bitOffset, numBits))
                    sets.append(setBitsFn(msg, field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return gets+sets

def enums(e):
    ret = ""
    for enum in e:
        keys = ""
        for option in enum["Options"]:
            keys += "    " + str(option["Value"]) + ",\n"
        keys = keys[:-2]
        values = ""
        for option in enum["Options"]:
            values += "    '" + option["Name"] + "',\n"
        values = values[:-2]
        # forward enum
        fwd = enum["Name"]+"Enum = containers.Map({...\n"
        fwd += keys
        fwd += "}, {\n"
        fwd += values
        fwd += "});\n"

        # Reverse enum
        back = "Reverse" + enum["Name"]+"Enum = containers.Map({...\n"
        back += values
        back += "}, {\n"
        back += keys
        back += "});\n"

        ret += fwd + back
    return ret

def declarations(msg, msg_enums):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(matlabFieldName(msg,field) + ";")
            if "Enum" in field:
                ret.append(matlabFieldName(msg,field) + "AsInt;")
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    ret.append(bits["Name"] + ";")
                    if "Enum" in bits:
                        ret.append(bits["Name"] + "AsInt;")
    return ret

def fieldDefault(field):
    if "Default" in field:
        return str(field["Default"])
    # should be based on type
    return "0"

def initField(msg, field):
    defaultValue = fieldType(field) + "("+ fieldDefault(field) +")"
    if MsgParser.fieldCount(field) > 1:
        ret =  "for index="+str(MsgParser.fieldCount(field))+": -1: 1\n"
        ret += "    obj." + matlabFieldName(msg,field) + "(index) = " + defaultValue + ";\n"
        ret += "end"
        return ret
    else:
        return  "obj." + matlabFieldName(msg,field) + " = " + defaultValue + ";"

def initBitfield(field, bits):
    paramType = fieldType(field)
    if "Offset" in bits or "Scale" in bits:
        paramType = typeForScaledInt(bits)
    defaultValue = paramType + "("+ fieldDefault(bits) +")"
    return  "obj." + MsgParser.BitfieldName(field, bits) + " = " +defaultValue + ";"

def initCode(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInit = initField(msg, field)
            if fieldInit:
                ret.append(fieldInit)
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    bits = initBitfield(field, bits)
                    if bits:
                        ret.append(bits)

    return ret

def undefinedMsgId():
    return "-1"

def addShift(base, value, shiftValue):
    ret = value
    if base != "":
        ret = "bitshift("+base+", "+str(shiftValue)+")"+"+"+value
    return ret

def getMsgID(msg):
    ret = ""
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "IDBits" in field:
                numBits = field["IDBits"]
                if "Enum" in field:
                    pass
                getStr = "obj."+matlabFieldName(msg,field)
                if "Enum" in field:
                    getStr += "AsInt"
                getStr = "uint32("+getStr+")"
                ret =  addShift(ret, getStr, numBits)
            if "Bitfields" in field:
                for bitfield in field["Bitfields"]:
                    if "IDBits" in bitfield:
                        numBits = bitfield["IDBits"]
                        if "Enum" in bitfield:
                            pass
                        getStr = "obj."+BitfieldName(field, bitfield)
                        if "Enum" in bitfield:
                            getStr += "AsInt"
                        getStr = "uint32("+getStr+")"
                        ret =  addShift(ret, getStr, numBits)
    return ret
    
def setMsgID(msg):
    ret = ""
    numBits = 0
    if "Fields" in msg:
        for field in reversed(msg["Fields"]):
            if "IDBits" in field:
                if numBits != 0:
                    ret += "\nid = bitshift(id, -" + str(numBits)+")\n"
                numBits = field["IDBits"]
                setStr = "bitand(id, "+Mask(numBits)+")"
                #if "Enum" in field:
                #    setStr = field["Enum"]+"("+setStr+")"
                ret +=  "obj."+matlabFieldName(msg,field)+" = "+setStr
            if "Bitfields" in field:
                for bitfield in reversed(field["Bitfields"]):
                    if "IDBits" in bitfield:
                        if numBits != 0:
                            ret += "\nid = id >> " + str(numBits)+"\n"
                        numBits = bitfield["IDBits"]
                        setStr = "bitand(id, "+Mask(numBits)+")"
                        #if "Enum" in bitfield:
                        #    setStr = bitfield["Enum"]+"("+setStr+")"
                        ret +=  "obj."+bitfield["Name"]+" = "+setStr
    return ret

oneOutputFilePerMsg = True



from PyQt5 import QtGui
from PyQt5.Qsci import QsciScintilla, QsciLexerPython

class EditWindow(QsciScintilla):
    CRASH_MARKER_NUM = 7
    DEBUG_MARKER_NUM = 8
    EXEC_MARKER_NUM = 9

    def __init__(self, parent=None):
        super(EditWindow, self).__init__(parent)
        
        # Set the default font
        font = QtGui.QFont()
        font.setFamily('Courier')
        font.setFixedPitch(True)
        font.setPointSize(10)
        self.setFont(font)
        self.setMarginsFont(font)

        # Margin 0 is used for line numbers
        fontmetrics = QtGui.QFontMetrics(font)
        self.setMarginsFont(font)
        self.setMarginWidth(0, fontmetrics.width("00000") + 6)
        self.setMarginLineNumbers(0, True)
        self.setMarginsBackgroundColor(QtGui.QColor("#cccccc"))

        self.setIndentationsUseTabs(False)
        self.setIndentationWidth(4)

        # Clickable margin 1 for showing markers
        self.setMarginSensitivity(1, True)
        self.marginClicked.connect(self.on_margin_clicked)
        
        # debug marker
        self.markerDefine(QsciScintilla.Circle, self.DEBUG_MARKER_NUM)
        self.setMarkerBackgroundColor(QtGui.QColor("#1111ee"), self.DEBUG_MARKER_NUM)

        # execute marker
        self.markerDefine(QsciScintilla.Circle, self.EXEC_MARKER_NUM)
        self.setMarkerBackgroundColor(QtGui.QColor("#11ee11"), self.EXEC_MARKER_NUM)

        # crash marker
        self.markerDefine(QsciScintilla.Circle, self.CRASH_MARKER_NUM)
        self.setMarkerBackgroundColor(QtGui.QColor("#ee1111"), self.CRASH_MARKER_NUM)

        # Brace matching: enable for a brace immediately before or after
        # the current position
        #
        self.setBraceMatching(QsciScintilla.SloppyBraceMatch)

        # Current line visible with special background color
        self.setCaretLineVisible(True)
        self.setCaretLineBackgroundColor(QtGui.QColor("#ffe4e4"))

        # Set Python lexer
        lexer = QsciLexerPython()
        lexer.setDefaultFont(font)
        self.setLexer(lexer)

        text = bytearray(str.encode("Arial"))
        self.SendScintilla(QsciScintilla.SCI_STYLESETFONT, 1, text)

        # not too small
        self.setMinimumSize(600, 450)
        
        self.last_exec_line = 0

    def on_margin_clicked(self, nmargin, nline, modifiers):
        # Toggle debug marker for the line the margin was clicked on
        if self.markersAtLine(nline) != 0:
            self.markerDelete(nline, self.DEBUG_MARKER_NUM)
        else:
            self.markerAdd(nline, self.DEBUG_MARKER_NUM)
            
    def exited(self):
        markerBitmask = self.markersAtLine(self.last_exec_line)
        if (1 << self.EXEC_MARKER_NUM) & markerBitmask:
            self.markerDelete(self.last_exec_line, self.EXEC_MARKER_NUM)
        self.markerAdd(self.last_exec_line, self.CRASH_MARKER_NUM)

    def crashed(self):
        markerBitmask = self.markersAtLine(self.last_exec_line)
        if (1 << self.EXEC_MARKER_NUM) & markerBitmask:
            self.markerDelete(self.last_exec_line, self.EXEC_MARKER_NUM)
        self.markerAdd(self.last_exec_line, self.CRASH_MARKER_NUM)

    # function called to indicate we executed up to a line of code
    def ran_to_line(self, nline):
        # change line number to zero based indexing
        nline = nline - 1
        markerBitmask = self.markersAtLine(self.last_exec_line)
        if (1 << self.EXEC_MARKER_NUM) & markerBitmask:
            self.markerDelete(self.last_exec_line, self.EXEC_MARKER_NUM)
        if (1 << self.CRASH_MARKER_NUM) & markerBitmask:
            # We have to delete crash marker twice!  Dunno why!
            self.markerDelete(self.last_exec_line, self.CRASH_MARKER_NUM)
            self.markerDelete(self.last_exec_line, self.CRASH_MARKER_NUM)
        if nline >= 0:
            self.markerAdd(nline, self.EXEC_MARKER_NUM)
            self.last_exec_line = nline
    
    def has_breakpoint(self, nline):
        # change line number to zero based indexing
        nline = nline - 1
        markerBitmask = self.markersAtLine(self.last_exec_line)
        if (1 << self.DEBUG_MARKER_NUM) & markerBitmask:
            return True
        return False



import os
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtSerialPort import QSerialPortInfo

class SelectSerialportDialog(QtWidgets.QDialog):
    portChanged = QtCore.pyqtSignal(str)
    @staticmethod
    def naIfEmpty(value):
        if not value:
            return "N/A"
        return value
    @staticmethod
    def naIfEmptyHex(value):
        if not value:
            return "N/A"
        try:
            return hex(int(value))
        except ValueError:
            pass
        return value
    def serialPortAlias(self, info):
        for name in os.listdir("/dev/"):
            if name not in (os.curdir, os.pardir):
                full = os.path.join("/dev", name)
                if os.path.islink(full):
                    if info.portName() == os.readlink(full):
                        return full.replace("/dev/","")
        return info.portName()

    def __init__(self, parent=None):
        super(SelectSerialportDialog, self).__init__(parent)
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.setWindowTitle("Select a Port")
        
        self.resize(800, 200)

        self.portsList = QtWidgets.QTreeWidget()
        openButton = QtWidgets.QPushButton("Open")
        openButton.clicked.connect(self.openPort)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.portsList)
        layout.addWidget(openButton)
        self.setLayout(layout)

        tableHeader = ["Name", "Description", "Mfg", "Serial#", "Location", "VendorID", "ProductID"]
        self.portsList.setHeaderLabels(tableHeader)
        for info in QSerialPortInfo.availablePorts():
            list = []
            description = info.description()
            manufacturer = info.manufacturer()
            serialNumber = info.serialNumber()
            list.append(self.serialPortAlias(info))
            list.append(self.naIfEmpty(description))
            list.append(self.naIfEmpty(manufacturer))
            list.append(self.naIfEmpty(serialNumber))
            list.append(info.systemLocation())
            list.append(self.naIfEmptyHex(info.vendorIdentifier()))
            list.append(self.naIfEmptyHex(info.productIdentifier()))

            self.portsList.addTopLevelItem(QtWidgets.QTreeWidgetItem(None, list))
        for i in range(0, len(tableHeader)):
            self.portsList.resizeColumnToContents(i)
        
    def openPort(self):
        cur_item = self.portsList.currentItem()
        if cur_item is not None:
            name = cur_item.text(0)
            print('opening %s' % name)
            self.portChanged.emit(name)
            self.close()




#!/usr/bin/python3
import datetime
import os
import time
import traceback
import requests

from msgtools.lib.messaging import Messaging
import msgtools.console.client

from influxdb import InfluxDBClient
from influxdb.client import InfluxDBClientError

# this allows sending Messages as data to an InfluxDB database
class InfluxDBConnection:
    MAX_POINTS = 128
    PRINT_TIME_INTERVAL = 10.0
    def __init__(self, msgClient, dbname='messages', hostname='localhost', port=8086, username='root', password='root'):
        self.hostname = hostname
        self.port = port
        self.db = InfluxDBClient(host=hostname, port=port, username=username, password=password, database=dbname)
        self.msgClient = msgClient
        self.cookie = 0
        
        self.stats = DBStats()
        
        # History Data message has to have either just Data with interleaved value and time,
        # or be an array of structs which also interleaves value and time in memory, but
        # gives two functions to access them that each look like an array.
        self.dataInfo = Messaging.findFieldInfo(Messaging.Messages.Network.History.Data.fields, "Data")
        self.dataValueInfo = Messaging.findFieldInfo(Messaging.Messages.Network.History.Data.fields, "Data_Value")
        self.dataTimestampInfo = Messaging.findFieldInfo(Messaging.Messages.Network.History.Data.fields, "Data_Timestamp")
        if self.dataInfo != None:
            # multiply by 2 because the result includes a value and a timestamp
            self.dataResultSize = self.dataInfo.size *2
            self.dataResultOffset = self.dataInfo.offset
        else:
            # Size of a result includes size of the value and the timestamp.
            self.dataResultSize = self.dataValueInfo.size + self.dataTimestampInfo.size
            # This needs to be whichever element is first!
            self.dataResultOffset = min(self.dataValueInfo.offset,self.dataTimestampInfo.offset)
    
    @staticmethod
    def FormattedTime(floatTime):
        return str(int(floatTime * 1e9))

    def handle_message(self, msg):
        if msg.MsgName().startswith("Network"):
            if msg.MsgName() == "Network.History.GetData":
                self.handle_query(msg)
        else:
            self.store_message(msg)
    
    def handle_timeout(self):
        stats_str = self.stats.report_stats(time.time(), self.PRINT_TIME_INTERVAL)
        if stats_str:
            print(stats_str)
    
    def store_message(self, msg):
        try:
            timeVal = msg.hdr.GetTime()
            if timeVal == 0.0:
                timeVal = time.time()
            else:
                timeInfo = Messaging.findFieldInfo(msg.hdr.fields, "Time")
                maxTime = timeInfo.maxVal
                # if it's not big enough to be an absolute timestamp, give up on using it and just use current time
                if maxTime != "DBL_MAX" and (maxTime == 'FLT_MAX' or float(maxTime) <= 2**32):
                    raise AttributeError
                if timeInfo.units == "ms":
                    timeVal = timeVal * 1.0e-3
                elif timeInfo.units == "us":
                    timeVal = timeVal * 1.0e-6
                elif timeInfo.units == "ns":
                    timeVal = timeVal * 1.0e-9
            timeVal = datetime.datetime.fromtimestamp(timeVal, datetime.timezone.utc)
        except AttributeError:
            timeVal = datetime.datetime.now()

        dbJson = {
                "time": str(timeVal),
                "measurement": msg.MsgName(),
                'fields':  {},
                'tags': {}
            }
        for fieldInfo in msg.hdr.fields:
            if len(fieldInfo.bitfieldInfo) == 0:
                if fieldInfo.idbits == 0 and fieldInfo.name != "Time" and fieldInfo.name != "DataLength":
                    dbJson['tags'][fieldInfo.name] = Messaging.get(msg.hdr, fieldInfo)
            else:
                for bitInfo in fieldInfo.bitfieldInfo:
                    if bitInfo.idbits == 0 and bitInfo.name != "Time" and bitInfo.name != "DataLength":
                        dbJson['tags'][bitInfo.name] = Messaging.get(msg.hdr, bitInfo)
        
        msgClass = type(msg)
        for fieldInfo in msgClass.fields:
            if fieldInfo.count == 1:
                if len(fieldInfo.bitfieldInfo) == 0:
                    dbJson['fields'][fieldInfo.name] = Messaging.get(msg, fieldInfo)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        dbJson['fields'][bitInfo.name] = Messaging.get(msg, bitInfo)
            else:
                # flatten arrays
                for i in range(0,fieldInfo.count):
                    dbJson['fields'][fieldInfo.name+"_"+str(i)] = Messaging.get(msg, fieldInfo, i)

        # Can't store with no fields!  Add a boolean to indicate emptiness
        if len(dbJson['fields']) == 0:
            dbJson['fields']['[EMPTY]'] = True

        #print("Create a retention policy")
        #retention_policy = 'awesome_policy'
        #client.create_retention_policy(retention_policy, '3d', 3, default=True)
        
        # record how many messages and bytes came in
        self.stats.messages_in += 1
        self.stats.bytes_in += msg.hdr.SIZE + msg.hdr.GetDataLength()
        try:
            self.db.write_points([dbJson]) #, retention_policy=retention_policy)
            self.stats.messages_out += 1
            # we don't know exactly what gets written across the network to InfluxDB, but
            # the length of the string of the JSON should be pretty close.
            self.stats.bytes_out += len(str(dbJson))
        except requests.exceptions.ConnectionError:
            self.stats.connection_errors += 1
        except Exception as e:
            print(traceback.format_exc())
            print(e)
            #print(dbJson)
            #print(msg)
            sys.exit()
    
    def handle_query(self, msg):
        msgquery = msg.GetQuery()
        msgAndFieldName = msgquery.split(",")[0]
        msgname = msgAndFieldName.split(".")[0]
        fieldname = msgAndFieldName.split(".")[1]
        tagQuery = msgquery.split(",")[1]
        if tagQuery != "":
            tagQuery += " AND "
        
        start = msg.GetStartTime()
        end = msg.GetEndTime()
        averagingPeriod = msg.GetAveragingPeriod()

        if averagingPeriod != 0:
            dbquery = 'SELECT "' + fieldname + '" FROM "' + msgname + '"'
        else:
            dbquery = 'SELECT MEAN("' + fieldname + '") FROM "' + msgname + '"'
        dbquery += " WHERE " + tagQuery + "time > " + FormattedTime(start)
        if end != 0:
            dbquery += " AND time < " + FormattedTime(end)
        else:
            end = datetime.datetime.now()
        if averagingPeriod != 0:
            # limit averaging period such that we don't get too many data points
            if avg / (end - start) > MAX_POINTS:
                avg = MAX_POINTS * (end - start)
            dbquery += " GROUP BY *,time("+FormattedTime(avg)+")"

        dbquery += " LIMIT " + str(MAX_POINTS)

        # send query to database, get results
        db_results = self.db.query(dbquery)

        # output results
        resultMsg = Messaging.Messages.Network.History.Result()
        resultMsg.SetCookie(self.cookie)
        resultCount = f(db_results)
        resultMsg.SetResultCount(resultCount)
        resultMsg.SetQuery(msgquery)
        self.msgClient.send_message(resultMsg)
        if self.dataInfo != None:
            maxDataPerMsg = self.dataInfo.count
        else:
            arrayOffset = self.dataTimestampInfo.offset
            arrayElemSize = self.dataValueInfo.size + dataTimestampInfo.size
            maxDataPerMsg = dataValueInfo.count
        resultMsgNumber = 0
        resultsLeftToSend = len(db_results)
        for result in db_results:
            dataMsg = Messaging.Messages.Network.History.Data()
            dataMsg.SetCookie(self.cookie)
            dataMsg.SetResultNumber(resultMsgNumber)
            dataRange = min(maxDataPerMsg, resultsLeftToSend)
            if self.dataInfo != None:
                for index in range(0, dataRange,2):
                    dataMsg.SetData(time(result), index)
                    dataMsg.SetData(value(result), index+1)
            else:
                for index in range(0, dataRange):
                    dataMsg.SetData_Timestamp(time(result), index)
                    dataMsg.SetData_Value(value(result), index)
            resultsLeftToSend -= dataRange
            resultMsgNumber += 1
            # truncate message buffer to actual amount of data
            msg_buffer_len = self.dataResultOffset + self.dataResultSize * dataRange
            self.msgClient.send_message(dataMsg)

        # Increment cookie so the next query's results are easy to tell apart from out query results
        # This is especially useful if two clients query at almost the same time (but only if they
        # have unique cookies!).
        self.cookie += 1

class DBStats:
    SUMMARY_FORMAT_STR = "{: >8}/{} msgs {: >8} bytes in {: >8} bytes out {: >8} cxn errors {: >8}/{} Msg/s {: >8} B/s in {: >8} B/s out {: >8} cxn errors/s"
    def __init__(self):
        self.messages_in = 0
        self.messages_out = 0
        self.bytes_in = 0
        self.bytes_out = 0
        self.connection_errors = 0
        self.last_time = 0
        self.last_messages_in = 0
        self.last_messages_out = 0
        self.last_bytes_in = 0
        self.last_bytes_out = 0
        self.last_connection_errors = 0
    
    def report_stats(self, now, time_interval):
        if now == None or now > self.last_time + time_interval:
            mips = (self.messages_in - self.last_messages_in)/time_interval
            mops = (self.messages_out - self.last_messages_out)/time_interval
            bips = (self.bytes_in - self.last_bytes_in)/time_interval
            bops = (self.bytes_out - self.last_bytes_out)/time_interval
            ceps = (self.connection_errors - self.last_connection_errors)/time_interval
            self.last_messages_in = self.messages_in
            self.last_messages_out = self.messages_out
            self.last_bytes_in = self.bytes_in
            self.last_bytes_out = self.bytes_out
            self.last_connection_errors = self.connection_errors
            self.last_time = now
            return self.SUMMARY_FORMAT_STR.format(
                self.messages_out, self.messages_in,
                self.bytes_in, self.bytes_out,
                self.connection_errors,
                mops, mips,
                bips, bops,
                ceps)
        return None

# this is client that reads from network, and writes to InfluxDB
class InfluxDBMsgClient:
    def __init__(self):
        Messaging.LoadAllMessages()

        self.connection = msgtools.console.client.Client("InfluxDB")
        
        self.db = InfluxDBConnection(self)
    
    def run(self):
        while True:
            # this blocks until message received, or timeout occurs
            timeout = 10.0 # value in seconds
            msg = self.connection.recv(timeout=timeout)
            if msg:
                self.db.handle_message(msg)
            self.db.handle_timeout()

# this is a CLI app that reads from network and writes to InfluxDB
def main(args=None):
    dbClient = InfluxDBMsgClient()
    try:
        dbClient.run()
    except KeyboardInterrupt:
        print('You pressed Ctrl+C!')
        dbClient.connection.stop()

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# TODO
#   add hotkey support (already reading options for it on cmdline)
#  Add annotation field.
#         text gets inserted in log whenever you hit enter in annotation box or click 'note' button next to annotate box.

import os
import sys
import argparse
from PyQt5 import QtCore, QtGui, QtWidgets
from time import strftime

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging
import msgtools.lib.gui
import msgtools.lib.txtreewidget

plottingLoaded=0
try:
    from msgtools.lib.msgplot import MsgPlot
    plottingLoaded=1
except ImportError as e:
    print("Error loading plot interface ["+str(e)+"]")
    print("Perhaps you forgot to install pyqtgraph.")
except RuntimeError as e:
    print("Error loading plot interface ["+str(e)+"]")
    print("Perhaps you need to install the PyQt5 version of pyqtgraph.")

DESCRIPTION='''
    MultiLog presents a UI that lets you control logging of messages on an 
    individual basis, and send and receive messages.  Pressing a button 
    starts/stops a log file.  Underscores in label or field are replaced 
    with spaces for display (to make entering command-line args easier)

    Log filenames will be composed like so:
        YEAR_MONTH_DAY.TEXT1.TEXT2.TAG1.log'''

EPILOG='''
Example usage:
    msgmultilog --note LABEL1 LABEL2 --button hotkey:X logtag:TAG1 label:LABEL3  
    --button hotkey:X logtag:TAG2 label:LABEL4 --show MSGNAME MSGNAME2
    --plot MSGNAME1=fieldname1,fieldname2 MSGNAME2 --plot MSGNAME3
    MSGNAME2=fieldname3,fieldname4 --send MSGNAME MSGNAME2

    --note: each --note argument adds a text field named the specified LABEL, and 
               the value of the text field will become part of the filename.  
               Example argument: LABEL1
    --annotation: creates a text field and "Add Note" button, such that when the button is clicked,
               the text from the text field is put into a Network.Note message.
    --button: each --button argument adds a pushbutton named for the specified LABEL, with a 
                hotkey of the specified key.  If logtag value is specified, a button press
                will start/stop logging, and the logtag will become part of the filename.
                If a send value is specified, a button press will send messages.
                Example argument: hotkey:X,logtag:TAG1,label:LABEL1
    --show: each --show argument adds a table view of that MSGNAME
    --row: switch to a row for active layout
    --endrow: end the current row of active layout
    --col: switch to a column for active layout
    --endcol: end the current column of active layout
    --plot: each --plot argument adds a new plot of the fields within MSGNAMEs that follow it.
                If fields left off, all fields within that MSGNAME are plotted.
                If multiple MSGNAME arguments are listed for one --plot, they will be added to
                the same plot.  Add [idx] to field name if array and want element
                other than element zero.  Example argument: MSGNAME=fieldname1,fieldname2[2].
                You can also specify ymin= and ymax= to set fixed vertical scaling.
    --timeslider: add a timeslider for all previously added plots.  subsequent plots will use a separate timeslider.
    --send:each --send argument is a message name that adds a tree view to edit a message with a 'send' 
                button to send it.  Example argument: MSGNAME
'''

local_args = {"--note":True, "--annotation": True, "--button": True, "--show":True, "--plot":True,
              "--send": True, "--tab": True, "--endtab": True, "--log": True, "--row": True,
              "--endrow": True, "--col": True, "--endcol": True, "--timeslider": True}
def base_args(argv):
    ret = []
    ignore = False
    for arg in argv:
        if ignore:
            if arg.startswith('--'):
                ignore = (arg in local_args) or (arg.startswith("--log="))
        else:
            ignore = (arg in local_args) or (arg.startswith("--log="))
        if not ignore:
            ret.append(arg)
    return ret
    
class Multilog(msgtools.lib.gui.Gui):
    def __init__(self, argv, parent=None):
        parser = argparse.ArgumentParser(description=DESCRIPTION, epilog=EPILOG, formatter_class=argparse.RawDescriptionHelpFormatter)

        parser = msgtools.lib.gui.Gui.addBaseArguments(parser)

        bargs = base_args(argv[1:])
        args=parser.parse_args(bargs)

        msgtools.lib.gui.Gui.__init__(self, "msgmultilog 0.1", args, parent)
        self.logFileType = 'json' # default to JSON
        
        # event-based way of getting messages
        self.RxMsg.connect(self.ProcessMessage)
        
        # handling of messages by sub-widgets
        self.msgHandlers = {}
        
        # Top level layout
        widget = QtWidgets.QWidget(self)
        mainLayout = QtWidgets.QVBoxLayout(widget)
        self.setCentralWidget(widget)
        self.statusMsg = QtWidgets.QLabel("NOT logging")
        mainLayout.addWidget(self.statusMsg)
        self.tabs = None

        # Layout stack handling
        self.layoutStack = []
        self.activeLayout = None
        mainLayout.type = 'main'
        self.pushLayout(mainLayout)
        
        self.hotKeys = {}
        self.noteFields = []
        self.buttons = []
        self.activeLogButton = None
        txMsgs = None
        firstPlot = None

        # Process command line args
        for i in range(1, len(argv)):
            argname = argv[i]

            argvalue = []
            i+=1
            while(1):
                if i >= len(argv) or argv[i].startswith("--"):
                    break
                argvalue.append(argv[i])
                i+=1
            if argname == '--note':
                hLayout = QtWidgets.QHBoxLayout()
                leftvLayout = QtWidgets.QVBoxLayout()
                rightvLayout = QtWidgets.QVBoxLayout()
                hLayout.addLayout(leftvLayout)
                hLayout.addLayout(rightvLayout)
                self.activeLayout.addLayout(hLayout)
                for field in argvalue:
                    label = field.replace("_", " ")

                    # add text field
                    leftvLayout.addWidget(QtWidgets.QLabel(label))
                    noteField = QtWidgets.QLineEdit()
                    rightvLayout.addWidget(noteField)
                    self.noteFields.append(noteField)
            elif argname == '--annotation':
                hLayout = QtWidgets.QHBoxLayout()
                self.activeLayout.addLayout(hLayout)
                self.annotation = QtWidgets.QLineEdit()
                add_note_button = QtWidgets.QPushButton("Add Note")
                add_note_button.clicked.connect(self.AddAnnotation)
                hLayout.addWidget(self.annotation)
                hLayout.addWidget(add_note_button)
            elif argname == '--button':
                options = {}
                for option in argvalue:
                    key, value = option.split(":")
                    if key in options:
                        options[key].append(value)
                    else:
                        options[key] = [value]

                # add button
                label = options["label"][0].replace("_", " ")
                button = QtWidgets.QPushButton(label)
                button.options = options
                # how to set hot key?  do it at window level, or at button level?  or something else?
                if 'hotkey' in options:
                    hotKey = options["hotkey"][0]
                    self.hotKeys[hotKey] = button
                button.label = label
                self.buttons.append(button)
                self.activeLayout.addWidget(button)
                
                button.clicked.connect(self.HandleButtonPress)

            elif argname == '--show':
                for msgname in argvalue:
                    subLayout = QtWidgets.QVBoxLayout()
                    self.activeLayout.addLayout(subLayout)
                    subLayout.addWidget(QtWidgets.QLabel(msgname))
                    msgClass = Messaging.MsgClassFromName[msgname]
                    msgWidget = msgtools.lib.gui.MsgTextWidget(msgClass)
                    subLayout.addWidget(msgWidget)
                    if not msgClass.ID in self.msgHandlers:
                        self.msgHandlers[msgClass.ID] = []
                    self.msgHandlers[msgClass.ID].append(msgWidget)

            elif argname == '--plot':
                if plottingLoaded:
                    thisPlot = None
                    multiple_messages = len(argvalue) != 1
                    yaxes = None
                    for plotarg in argvalue:
                        if "=" in plotarg:
                            split = plotarg.split("=")
                            msgname = split[0]
                            fieldNames = split[1].split(',')
                        else:
                            msgname = plotarg
                            fieldNames = []
                        # Check if they're setting vertical scale instead of a message to plot
                        if msgname == 'yaxes':
                            yaxes = [int(fieldNames[0]), int(fieldNames[1])]
                            if thisPlot:
                                thisPlot.setYAxis(yaxes)
                        else:
                            msgClass = Messaging.MsgClassFromName[msgname]
                            if firstPlot:
                                if thisPlot:
                                    # Add to existing plot
                                    for fieldName in fieldNames:
                                        thisPlot.addLine(msgClass, msgKey=None, fieldName=fieldName, fieldLabel=None, multiple_messages=multiple_messages)
                                        if not msgClass.ID in self.msgHandlers:
                                            self.msgHandlers[msgClass.ID] = []
                                        self.msgHandlers[msgClass.ID].append(thisPlot)
                                else:
                                    thisPlot = MsgPlot.plotFactory(self.newPlot, msgClass, fieldNames, **plotargs, multiple_messages=multiple_messages)
                                if yaxes:
                                    thisPlot.setYAxis(yaxes)
                            else:
                                thisPlot = MsgPlot.plotFactory(self.newPlot, msgClass, fieldNames, displayControls=False, multiple_messages=multiple_messages)
                                firstPlot = thisPlot
                                plotargs = {"runButton":firstPlot.runButton, "clearButton":firstPlot.clearButton, "timeSlider":firstPlot.timeSlider, "displayControls":False}

            elif argname == '--send':
                for msgname in argvalue:
                    msgClass = Messaging.MsgClassFromName[msgname]
                    if not txMsgs:
                        txMsgs = QtWidgets.QTreeWidget(parent)
                        txMsgs.setColumnCount(4)
                        txMsgsHeader = QtWidgets.QTreeWidgetItem(None, ["Message", "Field", "Value", "Units", "Description"])
                        txMsgs.setHeaderItem(txMsgsHeader)
                        self.activeLayout.addWidget(txMsgs)
                    
                    msg = msgClass()
                    # set fields to defaults
                    msgWidget = msgtools.lib.txtreewidget.MessageItem(editable=True, tree_widget=txMsgs, msg=msg, msg_key=None)
                    msgWidget.qobjectProxy.send_message.connect(self.on_tx_message_send)
            elif argname == '--row':
                l = QtWidgets.QHBoxLayout()
                l.type = 'row'
                self.activeLayout.addLayout(l)
                self.pushLayout(l)
            elif argname == '--col':
                l = QtWidgets.QVBoxLayout()
                l.type = 'col'
                self.activeLayout.addLayout(l)
                self.pushLayout(l)
            elif argname == '--tab':
                self.addTab(argvalue[0])
            elif argname == '--endrow':
                self.popLayout('row')
            elif argname == '--endcol':
                self.popLayout('col')
            elif argname == '--endtab':
                self.popLayout('tab')
            elif argname == '--log':
                self.logFileType = argvalue[0]
            elif argname.startswith('--log='):
                self.logFileType = argname.split("=")[1]
            elif argname == '--timeslider':
                if firstPlot:
                    self.addTimeSlider(firstPlot, self.activeLayout)
                firstPlot = None

        if firstPlot:
            self.addTimeSlider(firstPlot, mainLayout)

    def addTimeSlider(self, plot, layout):
        if plot:
            hLayout = QtWidgets.QHBoxLayout()
            layout.addLayout(hLayout)
            hLayout.addWidget(plot.runButton)
            hLayout.addWidget(plot.clearButton)
            hLayout.addWidget(QtWidgets.QLabel("Time Scale"))
            hLayout.addWidget(plot.timeSlider)
        
    def printLayout(self, prefix):
        layouts = prefix
        for l in self.layoutStack:
            layouts += " " + l.type
        print(layouts)

    def pushLayout(self, newLayout):
        self.layoutStack.append(newLayout)
        self.activeLayout = newLayout
        #self.printLayout('+')
    
    def popLayout(self, layoutType):
        #self.printLayout('-')
        found_it = False
        while len(self.layoutStack) > 1:
            p = self.layoutStack.pop()
            self.activeLayout = self.layoutStack[-1]
            if p.type == layoutType:
                found_it = True
                break
        if not found_it:
            print("Warning: Trying to remove layout %s, couldn't find it!" % layoutType)

    def addTab(self, name):
        if self.tabs == None:
            self.tabs = QtWidgets.QTabWidget()
            self.activeLayout.addWidget(self.tabs)
        tabWidget = QtWidgets.QWidget()
        self.tabs.addTab(tabWidget, name)
        vLayout = QtWidgets.QVBoxLayout()
        tabWidget.setLayout(vLayout)
        vLayout.type = 'tab'
        self.pushLayout(vLayout)

    def newPlot(self, plot):
        vLayout = QtWidgets.QVBoxLayout()
        vLayout.addWidget(QtWidgets.QLabel(plot.msgClass.MsgName()))
        vLayout.addWidget(plot)
        self.activeLayout.addLayout(vLayout)
        if not plot.msgClass.ID in self.msgHandlers:
            self.msgHandlers[plot.msgClass.ID] = []
        self.msgHandlers[plot.msgClass.ID].append(plot)
        
    def on_tx_message_send(self, msg):
        if not self.connected:
            self.OpenConnection()
        self.SendMsg(msg)

    def ProcessMessage(self, msg):
        if msg.ID in self.msgHandlers:
            handlersList = self.msgHandlers[msg.ID]
            for handler in handlersList:
                handler.addData(msg)

    def CreateLogFile(self, tag):
        self.CloseLogFile()

        # Start with a blank filename and add notes and a tag to it if they're defined
        filename = ""
        for noteField in self.noteFields:
            note = noteField.text()
            if note and note != "":
                print("added note [%s]" % (note))
                filename += note.replace(" ", "_") + "."
        if tag and tag != "":
            print("added tag [%s]" % (tag))
            filename += "." + tag

        currentDateTime = QtCore.QDateTime.currentDateTime()
        filename += currentDateTime.toString("yyyyMMdd-hhmmss") + "." + self.logFileType

        self.startLog(filename)
        self.statusMsg.setText("Logging to " + filename)

    def CloseLogFile(self):
        self.stopLog()
        self.statusMsg.setText("NOT Logging")

    def HandleButtonPress(self):
        # when a button is used to start a log, the text of that button changes to "Stop".
        # starting any other log will stop the current one (changing it's text back to normal)
        button = self.sender()
        if "logtag" in button.options:
            if button == self.activeLogButton:
                self.CloseLogFile()
                button.setText(button.label)
                self.activeLogButton = None
            else:
                if self.activeLogButton != None:
                    self.activeLogButton.setText(self.activeLogButton.label)
                self.CreateLogFile(button.options["logtag"][0])
                button.setText("Stop")
                self.activeLogButton = button
        if "send" in button.options:
            for opts in button.options['send']:
                msgname = opts.split("[")[0]
                fields = opts.split("[")[1].replace("]","").split(",")
                msgClass = Messaging.MsgClassFromName[msgname]
                msg = msgClass()
                for field in fields:
                    parts = field.split("=")
                    name = parts[0]
                    value = parts[1]
                    Messaging.set(msg, Messaging.findFieldInfo(msgClass.fields, name), value)
                self.SendMsg(msg)

    def AddAnnotation(self):
        annotation_msg = Messaging.Messages.Network.Note(Text=self.annotation.text())
        self.logMsg(annotation_msg)
        self.SendMsg(annotation_msg)

def main(args=None):
    app = QtWidgets.QApplication(sys.argv)
    msgApp = Multilog(sys.argv)
    msgApp.show()
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()



import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

# used as a prefix to namespace functions, enums, etc.
namespace = ""
enumClass = "class "
firstParam = ""
firstParamDecl = ""
const = " const"
enumNamespace = 0
functionPrefix = ""

def fieldType(field):
    typeStr = field["Type"]
    if "int" in typeStr:
        return typeStr + "_t"
    if typeStr == "float32":
        return "float";
    if typeStr == "float64":
        return "double";
    return "?"
    
def fnHdr(field):
    ret = "/* %s %s, (%s to %s)*/" % (MsgParser.fieldDescription(field), MsgParser.fieldUnits(field), MsgParser.fieldMin(field), MsgParser.fieldMax(field))
    return ret

def inline_min(x, y):
    return "((%s < %s) ? %s : %s)" % (x, y, x, y)

def arrayAccessor(field):
    # if we've got floating point conversion, don't do an array accessor
    if fieldHasConversion(field):
        return ""
    # if we're not an array, don't do an array accessor
    if MsgParser.fieldCount(field) == 1:
        return ""
    
    enumSetCast = ""
    enumGetCast = ""
    if "Enum" in field and namespace == "":
        enumGetCast = "("+fieldType(field)+")"
        enumSetCast = "("+namespace+field["Enum"]+")"
    # Functions to copy in and out.  These call the field Set/Get functions so
    # that they work regardless of endian, and for non-contiguous arrays.
    ret = '''\
%s
%svoid CopyIn%s(%s* in, int len)
{
    int count = %s;
    for(int i=0; i<count; i++)
    {
        %sSet%s(%s);
    }
}
''' % (fnHdr(field), functionPrefix, namespace+field["Name"], joinParams(firstParamDecl, "const "+fieldType(field)), inline_min("len", MsgParser.fieldCount(field)), namespace, field["Name"], joinParams(firstParam, enumSetCast+"in[i], i"))
    ret += '''\
%s
%svoid CopyOut%s(%s* out, int len)
{
    int count = %s;
    for(int i=0; i<count; i++)
    {
        out[i] = %sGet%s(%s);
    }
}
''' % (fnHdr(field), functionPrefix, namespace+field["Name"], joinParams(firstParamDecl, fieldType(field)), inline_min("len", MsgParser.fieldCount(field)), enumGetCast+namespace, field["Name"], joinParams(firstParam, "i"))

    # Only return a pointer to the array if we have natural alignment,
    # and are contiguous
    if(MsgParser.fieldLocation(field) % MsgParser.fieldSize(field) == 0 and
       MsgParser.fieldSize(field) == MsgParser.fieldArrayElementOffset(field)):
        loc = str(MsgParser.fieldLocation(field))
        access = "(%s*)&m_data[%s]" % (fieldType(field), loc)
        arrayAccessor = '''\
%s
%s* %s(%s)
{
    return %s;
}''' % (fnHdr(field), functionPrefix+fieldType(field), namespace+field["Name"], firstParamDecl, access)

        if MsgParser.fieldSize(field) != 1:
            if MsgParser.big_endian:
                arrayAccessor = "#if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__ \n" + arrayAccessor + "\n#endif\n"
            else:
                arrayAccessor = "#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__ \n" + arrayAccessor + "\n#endif\n"
        ret += arrayAccessor

    return ret

def castForScaledInt(field):
    ret = typeForScaledInt(field)
    if namespace != "":
        ret = "(" + ret + ")"
    return ret

def getFn(field, as_int=False):
    loc = str(MsgParser.fieldLocation(field))
    param = ""
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += "int idx"
    accessor_name = "Get"
    if MsgParser.nativeAlignment(field):
        accessor_name = "GetAligned"
    accessor_name += "BE" if MsgParser.big_endian else "LE"
    access = "%s_%s(&m_data[%s])" % (accessor_name, fieldType(field), loc)
    access = getMath(access, field, castForScaledInt(field), 'f', as_int=as_int)
    retType = fieldType(field)
    if fieldHasConversion(field) and not as_int:
        retType = typeForScaledInt(field)
    elif "Enum" in field and namespace == "":
        retType = namespace+field["Enum"]
        access = retType + "(" + access + ")"
    ret = '''\
%s
%s %s(%s)%s
{
    return %s;
}''' % (fnHdr(field), functionPrefix+retType, namespace+"Get"+field["Name"]+("AsInt" if as_int else ""), joinParams(firstParamDecl, param), const, access)
    if "float" in retType or "double" in retType:
        ret = "#ifndef DISABLE_FLOAT_ACCESSORS\n" + ret + "\n#endif\n"
    return ret

def setFn(field, as_int=False):
    paramType = fieldType(field)
    valueString = setMath("value", field, "("+fieldType(field)+")", 'f', as_int=as_int)
    if fieldHasConversion(field) and not as_int:
        paramType = typeForScaledInt(field)
    elif "Enum" in field and namespace == "":
        valueString = "("+paramType+")" + "(" + valueString + ")"
        paramType = namespace+field["Enum"]
    param = paramType + " value"
    loc = str(MsgParser.fieldLocation(field))
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += ", int idx"
    accessor_name = "Set"
    if MsgParser.nativeAlignment(field):
        accessor_name = "SetAligned"
    accessor_name += "BE" if MsgParser.big_endian else "LE"
    ret = '''\
%s
%s %s(%s)
{
    %s_%s(&m_data[%s], %s);
}''' % (fnHdr(field), functionPrefix+"void", namespace+"Set"+field["Name"]+("AsInt" if as_int else ""), joinParams(firstParamDecl, param), accessor_name, fieldType(field), loc, valueString)
    if "float" in paramType or "double" in paramType:
        ret = "#ifndef DISABLE_FLOAT_ACCESSORS\n" + ret + "\n#endif\n"
    return ret

def getBitsFn(field, bits, bitOffset, numBits, as_int=False):
    count = MsgParser.fieldCount(field)
    idx = "" if count == 1 else "idx"
    idxDecl = "" if count == 1 else "int idx"
    access = "(%sGet%s(%s) >> %s) & %s" % (namespace, field["Name"], joinParams(firstParam, idx), str(bitOffset), MsgParser.Mask(numBits))
    access = getMath(access, bits, castForScaledInt(bits), 'f', as_int=as_int)
    retType = fieldType(field)
    if fieldHasConversion(bits) and not as_int:
        retType = typeForScaledInt(bits)
    elif "Enum" in bits and namespace == "":
        retType = namespace+bits["Enum"]
        access = retType + "(" + access + ")"
    ret = '''\
%s
%s %s(%s)%s
{
    return %s;
}''' % (fnHdr(bits), functionPrefix+retType, namespace+"Get"+MsgParser.BitfieldName(field, bits)+("AsInt" if as_int else ""), joinParams(firstParamDecl, idxDecl), const, access)
    if "float" in retType or "double" in retType:
        ret = "#ifndef DISABLE_FLOAT_ACCESSORS\n" + ret + "\n#endif\n"
    return ret

def setBitsFn(field, bits, bitOffset, numBits, as_int=False):
    count = MsgParser.fieldCount(field)
    idx = "" if count == 1 else "idx"
    idxDecl = "" if count == 1 else "int idx"
    paramType = fieldType(field)
    valueString = setMath("value", bits, "("+fieldType(field)+")", 'f', as_int=as_int)
    if fieldHasConversion(bits) and not as_int:
        paramType = typeForScaledInt(bits)
    elif "Enum" in bits and namespace == "":
        valueString = "("+paramType+")" + "(" + valueString + ")"
        paramType = namespace+bits["Enum"]
    mask_bits = MsgParser.Mask(numBits)
    if numBits + bitOffset > 32:
        mask_bits += "ull"
    clear_bits = "%s << %s" % (mask_bits, str(bitOffset))
    oldVal = '''%s(%s) & ~(%s)''' % (namespace+"Get"+field["Name"], joinParams(firstParam, idx), clear_bits);
    newVal = '''(%s) | ((%s & %s) << %s)''' % (oldVal, valueString, MsgParser.Mask(numBits), str(bitOffset));
    ret = '''\
%s
%s %s(%s)
{
    %s(%s);
}''' % (fnHdr(bits), functionPrefix+"void", namespace+"Set"+MsgParser.BitfieldName(field, bits)+("AsInt" if as_int else ""), joinParams(firstParamDecl, paramType+" value", idxDecl), namespace+"Set"+field["Name"], joinParams(firstParam, newVal, idx))
    if "float" in paramType or "double" in paramType:
        ret = "#ifndef DISABLE_FLOAT_ACCESSORS\n" + ret + "\n#endif\n"
    return ret

def accessors(msg):
    gets = []
    sets = []
    arrayAccessors = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            gets.append(getFn(field))
            sets.append(setFn(field))
            if fieldHasConversion(field):
                gets.append(getFn(field, as_int=True))
                sets.append(setFn(field, as_int=True))
            arrAcc = arrayAccessor(field)
            if arrAcc != "":
                arrayAccessors.append(arrAcc)
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    gets.append(getBitsFn(field, bits, bitOffset, numBits))
                    sets.append(setBitsFn(field, bits, bitOffset, numBits))
                    if fieldHasConversion(bits):
                        gets.append(getBitsFn(field, bits, bitOffset, numBits, as_int=True))
                        sets.append(setBitsFn(field, bits, bitOffset, numBits, as_int=True))
                    bitOffset += numBits

    return gets+sets+arrayAccessors

def fieldDefault(field, as_enum=False):
    ret = field["Default"]
    if("Type" in field and "int" in field["Type"]) and ret > 2**31:
        ret = str(ret)+'u'
    if as_enum and "Enum" in field:
        retType = "<MSGFULLNAME>Message::"+field["Enum"]
        ret = "%s(%s)" % (retType, ret)
    return ret

def initField(field):
    if "Default" in field:
        if MsgParser.fieldCount(field) > 1:
            ret = "for (int i=0; i<" + str(MsgParser.fieldCount(field)) + "; i++)\n"
            ret += "    "+namespace+"Set" + field["Name"] + "(" + joinParams(firstParam, str(fieldDefault(field, True))) + ", i);" 
            return ret;
        else:
            return  namespace+"Set" + field["Name"] + "(" + joinParams(firstParam, str(fieldDefault(field, True))) + ");"
    return ""

def initBitfield(field, bits):
    if "Default" in bits:
        if MsgParser.fieldCount(field) > 1:
            ret = "for (int i=0; i<" + str(MsgParser.fieldCount(field)) + "; i++)\n"
            ret += "    "+namespace+"Set" + MsgParser.BitfieldName(field, bits) + "(" + joinParams(firstParam, str(fieldDefault(bits, True))) + ", i);" 
            return ret;
        else:
            return  namespace+"Set" + MsgParser.BitfieldName(field, bits) + "(" + joinParams(firstParam, str(fieldDefault(bits, True))) + ");"
    return ""

def initCode(msg):
    ret = []
    noInitCode = True
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInit = initField(field)
            if fieldInit:
                ret.append(fieldInit)
                noInitCode = False
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    bits = initBitfield(field, bits)
                    if bits:
                        ret.append(bits)
                        noInitCode = False
    if noInitCode:
        ret.append("UNUSED(m_data);")

    return ret

def enums(e):
    ret = ""
    for enum in e:
        ret +=  "enum " + enumClass + namespace + enum["Name"]+" {"
        for option in enum["Options"]:
            optionName = OptionName(option)
            if enumNamespace != 0:
                optionName = "<MSGFULLNAME>"+"_"+enum["Name"] + "_" + optionName
            ret += optionName+" = "+str(option["Value"]) + ', '
        ret = ret[:-2]
        ret += "};\n"
    return ret

def fieldReflectionType(field):
    ret = fieldType(field)
    if ret == "double" or ret == "float":
        return "FloatFieldInfo"

    if ret.startswith("int"):
        ret = "IntFieldInfo"
    if ret.startswith("uint"):
        ret = "UIntFieldInfo"

    if "NumBits" in field:
        ret = "BitfieldInfo"
        if fieldHasConversion(field):
            ret = "ScaledBitfieldInfo"
    else:
        if fieldHasConversion(field):
            ret = "ScaledFieldInfo"
    if "Enum" in field:
        ret = "EnumFieldInfo"
    return ret

def fieldReflectionBitsType(field, bits):
    ret = fieldType(field)
    if ret == "double" or ret == "float":
        return "FloatFieldInfo"

    if ret.startswith("int"):
        ret = "IntFieldInfo"
    if ret.startswith("uint"):
        ret = "UIntFieldInfo"

    if "NumBits" in bits:
        ret = "BitfieldInfo"
        if fieldHasConversion(bits):
            ret = "ScaledBitfieldInfo"
    else:
        if fieldHasConversion(field):
            ret = "ScaledFieldInfo"
    if "Enum" in field:
        ret = "EnumFieldInfo"
    return ret

def fieldReflection(field):
    loc = str(MsgParser.fieldLocation(field))
    type = fieldReflectionType(field)
    params = type;
    params += "("
    params += '"'+field["Name"] + '"'
    params += ', "' + MsgParser.fieldDescription(field) + '"'
    params += ', "' + MsgParser.fieldUnits(field) + '"'
    params += ", " + loc
    params += ", " + str(MsgParser.fieldSize(field))
    params += ", " + str(MsgParser.fieldCount(field))
    if fieldHasConversion(field):
        if "Scale" in field:
            params += ", " + str(field["Scale"])
        else:
            params += ", 1.0"
        if "Offset" in field:
            params += ", " + str(field["Offset"])
        else:
            params += ", 0.0"
    params += ")"
    return params

def fieldBitsReflection(field, bits, bitOffset, numBits):
    loc = str(MsgParser.fieldLocation(field))
    type = fieldReflectionBitsType(field, bits)
    params = type;
    params += "("
    params += '"'+bits["Name"] + '"'
    params += ', "' + MsgParser.fieldDescription(bits) + '"'
    params += ', "' + MsgParser.fieldUnits(bits) + '"'
    params += ", " + loc
    params += ", " + str(MsgParser.fieldSize(field))
    params += ", " + str(MsgParser.fieldCount(bits))
    if fieldHasConversion(bits):
        if "Scale" in bits:
            params += ", " + str(bits["Scale"])
        else:
            params += ", 1.0"
        if "Offset" in bits:
            params += ", " + str(bits["Offset"])
        else:
            params += ", 0.0"
    params += ", " + str(bitOffset)
    params += ", " + str(numBits)
    params += ")"
    return params

def reflection(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldReflection(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsReflection(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)

def fieldMin(field):
    ret = str(MsgParser.fieldMin(field))
    if fieldHasConversion(field):
        ret += 'f'
    else:
        if fieldIsInt(field) and not fieldIsSigned(field):
            ret += 'U'
    return ret

def fieldMax(field):
    ret = str(MsgParser.fieldMax(field))
    if fieldHasConversion(field):
        ret += 'f'
    else:
        if fieldIsInt(field) and not fieldIsSigned(field):
            ret += 'U'
    return ret

def genericInfo(field, loc, type):
    params  = '    int static constexpr loc   = ' + loc + ';\n'
    params += '    '+type+' static constexpr max   = ' + fieldMax(field) + ';\n'
    params += '    '+type+' static constexpr min   = ' + fieldMin(field) + ';\n'
    params += '    char static constexpr units[] = "' + str(MsgParser.fieldUnits(field)) + '"' + ';\n'
    params += '    int static constexpr count = ' + str(MsgParser.fieldCount(field)) + ';\n'
    if "Default" in field:
        params += '    '+type+' static constexpr defaultValue = ' + str(fieldDefault(field)) + ";\n" 
    if "Scale" in field:
        params += '    auto static constexpr scale = ' + str(field["Scale"]) + ';\n'
    if "Offset" in field:
        params += '    auto static constexpr offset = ' + str(field["Offset"]) + ';\n'
    return params
    
def fieldInfo(field):
    retType = fieldType(field)
    if fieldHasConversion(field):
        retType = typeForScaledInt(field)

    params  = 'struct ' + field["Name"] + 'FieldInfo {\n'
    params += genericInfo(field, str(MsgParser.fieldLocation(field)), retType)
    params += '};\n'
    return params

def fieldBitsInfo(field, bits, bitOffset, numBits):
    retType = fieldType(field)
    if fieldHasConversion(bits):
        retType = typeForScaledInt(bits)

    params  = 'struct ' + bits["Name"] + 'FieldInfo {\n'
    params += genericInfo(bits, str(MsgParser.fieldLocation(field)), retType)
    params += '    auto static constexpr bitOffset = ' + str(bitOffset) + ';\n'
    params += '    auto static constexpr numBits   = ' + str(numBits) + ';\n'
    params += '};\n'
    return params

def fieldInfos(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldInfo(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsInfo(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)

def structUnpacking(msg):
    ret = []

    obj_name = namespace
    struct_name = "s->"
    if firstParam == "":
        obj_name = "msg."
        struct_name = ""

    if "Fields" in msg:    
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    ret.append(struct_name+bits["Name"] + " = "+obj_name+"Get" + bits["Name"] + "(%s);" % (firstParam))
            else:
                if MsgParser.fieldCount(field) == 1:
                    ret.append(struct_name+field["Name"] + " = "+obj_name+"Get" + field["Name"] + "(%s);" % (firstParam))
                else:
                    ret.append("for(int i=0; i<"+str(MsgParser.fieldCount(field))+"; i++)")
                    ret.append("    "+struct_name+field["Name"] + "[i] = "+obj_name+"Get" + field["Name"] + "(%s);" % (joinParams(firstParam, "i")))
            
    return "\n".join(ret)
    
def structPacking(msg):
    ret = []
    
    obj_name = namespace
    struct_name = "s->"
    if firstParam == "":
        obj_name = "msg."
        struct_name = ""

    if "Fields" in msg:    
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    ret.append(obj_name+"Set" + bits["Name"] + "(%s);" % (joinParams(firstParam, struct_name+bits["Name"])))
            else:
                if MsgParser.fieldCount(field) == 1:
                    ret.append(obj_name+"Set" + field["Name"] + "(%s);" % (joinParams(firstParam, struct_name+field["Name"])))
                else:
                    ret.append("for(int i=0; i<"+str(MsgParser.fieldCount(field))+"; i++)")
                    ret.append("    "+obj_name+"Set" + field["Name"] + "(%s);" % (joinParams(firstParam,struct_name+field["Name"] + "[i], i")))
                
    return "\n".join(ret)
    
def declarations(msg, msg_enums):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    retType = fieldType(field)
                    if fieldHasConversion(bits):
                        retType = typeForScaledInt(bits)
                    elif "Enum" in bits:
                        retType = "<MSGFULLNAME>Message::"+bits["Enum"]
                    ret.append(retType + " " + bits["Name"] + ";")
            else:
                retType = fieldType(field)
                if fieldHasConversion(field):
                    retType = typeForScaledInt(field)
                elif "Enum" in field:
                    retType = "<MSGFULLNAME>Message::"+field["Enum"]
                if MsgParser.fieldCount(field) == 1:
                    ret.append(retType + " " + field["Name"] + ";")
                else:
                    ret.append(retType + " " + field["Name"] + "["+str(MsgParser.fieldCount(field))+"];")
    return ret

def getMsgID(msg):
    return baseGetMsgID("", "", "CAST_ENUMS", 0, msg)
    
def setMsgID(msg):
    return baseSetMsgID("", "", 1, 0, msg)



import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

# https://www.html5rocks.com/en/tutorials/webgl/typed_arrays/
# https://github.com/kig/DataStream.js
# https://www.html5rocks.com/en/tutorials/websockets/basics/
def endian_string():
    if MsgParser.big_endian:
        return ''
    else:
        return ', true'

def fieldType(field):
    fieldTypeDict = \
    {"uint64":"long", "uint32":"long","uint16": "int",   "uint8": "short",
     "int64":"long",   "int32":"int",  "int16": "short",  "int8": "char",
      "float64":"double", "float32":"float"}
    typeStr = field["Type"]
    return field["Type"].capitalize()

def fnHdr(field):
    ret = "// %s %s, (%s to %s)" % (MsgParser.fieldDescription(field), MsgParser.fieldUnits(field), MsgParser.fieldMin(field), MsgParser.fieldMax(field))
    return ret

def typeForScaledInt(field):
    numBits = MsgParser.fieldNumBits(field)
    if numBits > 24:
        return "double"
    return "float"

def enumLookup(field):
    lookup  = "if(value in <MSGNAME>."+ str(field["Enum"])+")\n"
    lookup += "        value = <MSGNAME>."+ str(field["Enum"])+"[value];\n"
    lookup += "    "
    return lookup

def reverseEnumLookup(field):
    lookup = "if(!enumAsInt)\n"
    lookup += "    if(value in <MSGNAME>.Reverse"+ str(field["Enum"])+")\n"
    lookup += "        value = <MSGNAME>.Reverse" + str(field["Enum"]) + "[value];\n    "
    return lookup

def getFn(field):
    loc = str(MsgParser.fieldLocation(field))
    param = ""
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += "idx"
    if "Enum" in field:
        if param != "":
            param += ", "
        param += "enumAsInt=false"
    access = "(this.m_data.get%s(%s%s))" % (fieldType(field), loc, endian_string())
    access = getMath(access, field, "")
    cleanup = ""
    if "Enum" in field:
        cleanup = reverseEnumLookup(field)
    ret = '''\
%s
<MSGNAME>.prototype.Get%s = function(%s)
{
    var value = %s;
    %sreturn value;
};''' % (fnHdr(field), field["Name"], param, access, cleanup)
    if MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
        ret += '''
%s
<MSGNAME>.prototype.Get%sString = function()
{
    var value = '';
    for(i=0; i<%s && i<this.hdr.GetDataLength()-%s; i++)
    {
        nextChar = String.fromCharCode(this.Get%s(i));
        if(nextChar == '\\0')
            break;
        value += nextChar;
    }
    return value;
};''' % (fnHdr(field), field["Name"], str(MsgParser.fieldCount(field)), str(MsgParser.fieldLocation(field)), field["Name"])
    return ret

def setFn(field):
    valueString = setMath("value", field, "")
    lookup = ""
    if "Enum" in field:
        # find index that corresponds to string input param
        lookup = enumLookup(field)        
    param = "value"
    loc = str(MsgParser.fieldLocation(field))
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += ", idx"
    ret = '''\
%s
<MSGNAME>.prototype.Set%s = function(%s)
{
    %sthis.m_data.set%s(%s, %s%s);
};''' % (fnHdr(field), field["Name"], param, lookup, fieldType(field), loc, valueString, endian_string())
    if MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
        ret += '''
%s
<MSGNAME>.prototype.Set%sString = function(value)
{
    for(i=0; i<%s && i<value.length; i++)
    {
        this.Set%s(value[i].charCodeAt(0), i);
    }
};''' % (fnHdr(field), field["Name"], str(MsgParser.fieldCount(field)), field["Name"])
    return ret

def getBitsFn(field, bits, bitOffset, numBits):
    idx = "" if MsgParser.fieldCount(field) == 1 else "idx"
    access = "(this.Get%s(%s) / 2**%s) & %s" % (field["Name"], idx, str(bitOffset), MsgParser.Mask(numBits))
    access = getMath(access, bits, "")
    enumParam = "enumAsInt=false" if "Enum" in bits else ""
    cleanup = ""
    if "Enum" in bits:
        cleanup = reverseEnumLookup(bits)
    ret = '''\
%s
<MSGNAME>.prototype.Get%s = function(%s)
{
    var value = %s;
    %sreturn value;
};''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), joinParams(idx, enumParam), access, cleanup)
    return ret

def setBitsFn(field, bits, bitOffset, numBits):
    valueString = setMath("value", bits, "")
    lookup = ""
    if "Enum" in bits:
        # find index that corresponds to string input param
        lookup = enumLookup(bits)
    idx = "" if MsgParser.fieldCount(field) == 1 else "idx"
    set_value = "(this.Get%s() & ~(%s * 2**%s)) | ((%s & %s) * 2**%s)" % (field["Name"], MsgParser.Mask(numBits), str(bitOffset), valueString, MsgParser.Mask(numBits), str(bitOffset))
    ret = '''\
%s
<MSGNAME>.prototype.Set%s = function(%s)
{
    %sthis.Set%s(%s);
};''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), joinParams("value", idx), lookup, field["Name"], joinParams(set_value, idx))
    return ret

def accessors(msg):
    gets = []
    sets = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            gets.append(getFn(field))
            sets.append(setFn(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    gets.append(getBitsFn(field, bits, bitOffset, numBits))
                    sets.append(setBitsFn(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return gets+sets

def initField(field):
    if "Default" in field:
        if MsgParser.fieldCount(field) > 1:
            ret = "for (i=0; i<" + str(MsgParser.fieldCount(field)) + "; i++)\n"
            ret += "    this.Set" + field["Name"] + "(" + str(field["Default"]) + ", i);" 
            return ret;
        else:
            return  "this.Set" + field["Name"] + "(" + str(field["Default"]) + ");"
    return ""

def initBitfield(field, bits):
    if "Default" in bits:
        if MsgParser.fieldCount(field) > 1:
            ret = "for (i=0; i<" + str(MsgParser.fieldCount(field)) + "; i++)\n"
            ret += "    this.Set" + MsgParser.BitfieldName(field, bits) + "(" +str(bits["Default"]) + ", i);"
            return ret;
        else:
            return  "this.Set" + MsgParser.BitfieldName(field, bits) + "(" +str(bits["Default"]) + ");"
    return ""

def initCode(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInit = initField(field)
            if fieldInit:
                ret.append(fieldInit)
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    bits = initBitfield(field, bits)
                    if bits:
                        ret.append(bits)

    return ret

def enums(e):
    ret = ""
    for enum in e:
        # forward enum
        fwd = "<MSGNAME>." + enum["Name"]+" = {};\n"
        for option in enum["Options"]:
            fwd += "<MSGNAME>."+enum["Name"] + "[\"" + str(option["Name"]) + "\"] = "+str(option["Value"])+";\n"

        # Reverse enum
        back = "<MSGNAME>.Reverse" + enum["Name"]+" = {};\n"
        back += "for(key in <MSGNAME>."+enum["Name"]+") {\n"
        back += "    <MSGNAME>.Reverse" + enum["Name"] + "[<MSGNAME>."+enum["Name"]+"[key]" +"] = key;\n"
        back += "}\n"

        ret += fwd + back
    return ret

def reflectionInterfaceType(field):
    type = field["Type"]
    if "float" in type or "Offset" in field or "Scale" in field:
        type = "float"
    elif MsgParser.fieldUnits(field) == "ASCII":
        type = "string"
    elif "Enum" in field:
        type = "enumeration"
    else:
        type = "int"
    return type

def bitsReflectionInterfaceType(field):
    type = "int"
    if "Offset" in field or "Scale" in field:
        type = "float"
    elif MsgParser.fieldUnits(field) == "ASCII":
        type = "string"
    elif "Enum" in field:
        type = "enumeration"
    else:
        type = "int"
    return type

def bitfieldReflection(msg, field, bits):
    name = bits["Name"]
    ret = "{"+\
              'name:"'+name + '",'+\
              'type:"'+bitsReflectionInterfaceType(bits) + '",'+\
              'units:"'+MsgParser.fieldUnits(bits) + '",'+\
              'minVal:"'+str(MsgParser.fieldMin(bits)) + '",'+\
              'maxVal:"'+str(MsgParser.fieldMax(bits)) + '",'+\
              'description:"'+MsgParser.fieldDescription(bits) + '",'+\
              'get:"Get' + name + '",'+\
              'set:"Set' + name  + '", '
    if "Enum" in bits:
        ret += "enumLookup : [<MSGNAME>."+  bits["Enum"]+", " + "<MSGNAME>.Reverse" + bits["Enum"]+"]}"
    else:
        ret += "enumLookup : []}"
    return ret

def fieldReflection(msg, field):
    fieldFnName = field["Name"]
    fieldCount = MsgParser.fieldCount(field)
    if MsgParser.fieldCount(field) != 1 and MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
        fieldFnName = field["Name"]+"String"
        fieldCount = 1
    fieldInfo = "{"+\
                  'name:"'+field["Name"] + '",'+\
                  'type:"'+reflectionInterfaceType(field) + '",'+\
                  'units:"'+MsgParser.fieldUnits(field) + '",'+\
                  'minVal:"'+str(MsgParser.fieldMin(field)) + '",'+\
                  'maxVal:"'+str(MsgParser.fieldMax(field)) + '",'+\
                  'description:"'+MsgParser.fieldDescription(field) + '",'+\
                  'get:"Get' + fieldFnName + '",'+\
                  'set:"Set' + fieldFnName  + '",'+\
                  'count:'+str(fieldCount) + ', '
    if "Bitfields" in field:
        bitfieldInfo = []
        for bits in field["Bitfields"]:
            bitfieldInfo.append("    " + bitfieldReflection(msg, field, bits))
        fieldInfo += "bitfieldInfo : [\n" + ",\n".join(bitfieldInfo) + "], "
    else:
        fieldInfo += "bitfieldInfo : [], "
    if "Enum" in field:
        fieldInfo += "enumLookup : [<MSGNAME>." + field["Enum"]+", " + "<MSGNAME>.Reverse" + field["Enum"]+"]}"
    else:
        fieldInfo += "enumLookup : []}"
    return fieldInfo

def reflection(msg):
    fieldInfos = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInfos.append(fieldReflection(msg, field))
    return ",\n".join(fieldInfos)

def genericInfo(field, loc, type):
    params  = '    static final int loc   = ' + loc + ';\n'
    params += '    static final '+type+' max   = ' + fieldMax(field) + ';\n'
    params += '    static final '+type+' min   = ' + fieldMin(field) + ';\n'
    params += '    static final String units = "' + str(MsgParser.fieldUnits(field)) + '"' + ';\n'
    params += '    static final int count = ' + str(MsgParser.fieldCount(field)) + ';\n'
    if "Default" in field:
        params += '    static final '+type+' defaultValue = ' + str(field["Default"]) + ";\n" 
    if "Scale" in field:
        params += '    static final int scale = ' + str(field["Scale"]) + ';\n'
    if "Offset" in field:
        params += '    static final int offset = ' + str(field["Offset"]) + ';\n'
    return params
    
def fieldInfo(field):
    retType = fieldType(field)
    if "Offset" in field or "Scale" in field:
        retType = typeForScaledInt(field)

    params  = 'class ' + field["Name"] + 'FieldInfo {\n'
    params += genericInfo(field, str(MsgParser.fieldLocation(field)), retType)
    params += '};\n'
    return params

def fieldBitsInfo(field, bits, bitOffset, numBits):
    retType = fieldType(field)
    if "Offset" in bits or "Scale" in bits:
        retType = typeForScaledInt(bits)

    params  = 'class ' + bits["Name"] + 'FieldInfo {\n'
    params += genericInfo(bits, str(MsgParser.fieldLocation(field)), retType)
    params += '    static final int bitOffset = ' + str(bitOffset) + ';\n'
    params += '    static final int numBits   = ' + str(numBits) + ';\n'
    params += '};\n'
    return params

def fieldInfos(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldInfo(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsInfo(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)
    
def declarations(msg, msg_enums):
    return []

def getMsgID(msg):
    return baseGetMsgID("this.", "", 0, 1, msg).replace("<<", " * 2**")
    
def setMsgID(msg):
    return baseSetMsgID("this.", "", 0, 1, msg).replace(">>", " / 2**")

def structUnpacking(msg):
    ret = []

    if "Fields" in msg:    
        for field in msg["Fields"]:
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    ret.append('try { ret["'+bits["Name"] + '"] = this.Get' + bits["Name"] + "(); } catch (err) {}")
            else:
                if MsgParser.fieldCount(field) == 1:
                    ret.append('try { ret["'+field["Name"] + '"] = this.Get' + field["Name"] + "(); } catch (err) {}")
                else:
                    if MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
                        ret.append('try { ret["'+field["Name"] + '"] = this.Get' + field["Name"] + "String(); } catch (err) {}")
                    else:
                        ret.append('try { ret["'+field["Name"] + '"] = []; } catch (err) {}')
                        ret.append('try { ')
                        ret.append("    for(i=0; i<"+str(MsgParser.fieldCount(field))+"; i++)")
                        ret.append('        ret["'+field["Name"] + '"][i] = this.Get' + field["Name"] + "(i);")
                        ret.append('} catch (err) {}')
            
    return "\n".join(ret)



from msgtools.lib.message import Message
from msgtools.lib.messaging import Messaging
from msgtools.lib.header_translator import HeaderHelper, HeaderTranslator
import ctypes
import importlib
import json
import math

class MessageFileReader:
    def __init__(self):
        # to handle timestamp wrapping
        self._timestampOffset = 0
        self._lastTimestamp = 0

    def error_fn(self, str):
        print(str)

    def read_file(self, filename, header_name, ignore_invalid=False):
        # load the messages if not already loaded
        if Messaging.hdr == None:
            try:
                if header_name != None:
                    Messaging.LoadAllMessages(headerName=header_name)
                else:
                    Messaging.LoadAllMessages()
            except RuntimeError as e:
                print(e)
                quit()
            except:
                import traceback
                print(traceback.format_exc())
                quit()

        # find the specific header used in the log, and create a translator
        # between that and regular messages.
        if header_name != None and header_name != Messaging.hdr.__name__:
            header_module = importlib.import_module("headers." + header_name)
            self.log_header = getattr(header_module, header_name)
            self.header_translator = HeaderTranslator(self.log_header, Messaging.hdr)
        else:
            self.log_header = Messaging.hdr
            self.header_translator = None
        self.header_helper = HeaderHelper(self.log_header, self.error_fn)

        if filename.endswith(".json"):
            self.read_json_file(filename, ignore_invalid)
        else:
            self.read_binary_file(filename)

    def read_binary_file(self, filename):
        with open(filename, mode='rb') as f:
            while(1):
                # If we need to translate the header, then read the header and body separately.
                # Otherwise do a single big read and seek backwards for the next message's start, because that is ~10% faster.
                if self.header_translator:
                    # read bytes, create a header
                    msg_bytes = f.read(self.log_header.SIZE)
                    if len(msg_bytes) < self.log_header.SIZE:
                        break
                    hdr = self.log_header(msg_bytes)

                    if not self.header_helper.header_valid(hdr):
                        print("Invalid header!")
                        break

                    # read the body
                    msg_body = f.read(hdr.GetDataLength())

                    #if not self.header_helper.body_valid(hdr, msg.rawBuffer()[self.log_header.SIZE:]):
                    if not self.header_helper.body_valid(hdr, msg_body):
                        print("Invalid body!")
                        break

                    network_msg = self.header_translator.translateHdrAndBody(hdr, msg_body)
                else:
                    # <- read a kB of data so we do fewer (but larger) reads
                    last_header_pos = f.tell()
                    msg_bytes = ctypes.create_string_buffer(Messaging.hdr.SIZE+1024)
                    length_read = f.readinto(msg_bytes)
                    # Construct a header object
                    hdr = Messaging.hdr(msg_bytes)

                    if length_read < Messaging.hdr.SIZE:
                        break

                    if not self.header_helper.header_valid(hdr):
                        print("Invalid header!")
                        break

                    # <- seek back to just after the last message's data (where the next header starts)
                    f.seek(last_header_pos+Messaging.hdr.SIZE+hdr.GetDataLength())

                    hdr = Messaging.hdr(msg_bytes)

                    if not self.header_helper.body_valid(hdr, hdr.rawBuffer()[Messaging.hdr.SIZE:]):
                        print("Invalid body!")
                        break
                    
                    network_msg = hdr
                
                # Get a specifically typed message, to give to process_msg
                msg = Messaging.MsgFactory(network_msg)
                self.process_message(msg)

    def read_json_file(self, filename, ignore_invalid=False):
        with open(filename) as f:
            for line in f:
                msg = Message.fromJson(line, ignore_invalid)
                self.process_message(msg)
